"""Provides the Calendar and CalendarDay utility classes.

The :class:`pyrealm.core.calendar.Calendar` class is currently used to support the
operation of the :mod:`~pyrealm.splash` submodule, which requires the Julian day, number
of days in the year and year for solar calculations. The class provides iterable and
indexable access to a date sequence for use in those calculations, returning individual
:class:`pyrealm.core.calendar.CalendarDay` instances.

It is possible that this could be replaced with xarray dt accessors if pyrealm adopts
xarray data structures.
"""

from collections.abc import Generator, Sized
from dataclasses import dataclass, field
from typing import Any

import numpy as np
from numpy.typing import NDArray


@dataclass
class CalendarDay:
    """The CalendarDay class.

    This dataclass holds a :class:`numpy.datetime64` datetime representing a day and the
    corresponding year, julian day (the integer index of the day within the year)
    and the total number of days in the year.
    """

    date: np.datetime64
    """The date of the instance as numpy.datetime64."""
    year: int
    """The year of the instance as an integer."""
    julian_day: int
    """The julian day of the instance as an integer."""
    days_in_year: int
    """The total number of days in the year."""

    def __repr__(self) -> str:
        """A custom representation method.

        The dataclass implementation does not support custom field representation, so
        this replaces the default.
        """
        return (
            f"CalendarDay(date={self.date!s}, year={self.year}, "
            f"julian_day={self.julian_day}, days_in_year={self.days_in_year})"
        )


@dataclass
class Calendar(Sized):
    """The Calendar class.

    This utility class takes a numpy array of :class:`numpy.datetime64` values
    representing a time series of individual days and calculates the date, year, julian
    day and days in the year for each observation. The class object can be iterated
    over, yielding each date in turn, and dates within the series can also be accessed
    by index. In both cases, the returned object is a
    :class:`~pyrealm.core.calendar.CalendarDay` instance.

    Examples:
        >>> days=np.arange(
        ...     np.datetime64("2000-01-01"),
        ...     np.datetime64("2002-01-01"),
        ...     np.timedelta64(1, "D")
        ... )
        >>> cal = Calendar(days)
        >>> cal
        Calendar(2000-01-01, 2001-12-31)
        >>> len(cal) == (366 + 365)
        True
        >>> cal[0]
        CalendarDay(date=2000-01-01, year=2000, julian_day=1, days_in_year=366)
        >>> for date in cal:
        ...     pass
        >>> date
        CalendarDay(date=2001-12-31, year=2001, julian_day=365, days_in_year=365)
    """

    dates: NDArray[np.datetime64]
    """A numpy array containing :class:`numpy.datetime64` values."""
    year: NDArray[np.int_] = field(init=False)
    """A numpy array giving the year of each datetime."""
    julian_day: NDArray[np.int_] = field(init=False)
    """A numpy array giving the julian day of each datetime."""
    days_in_year: NDArray[np.int_] = field(init=False)
    """A numpy array giving the number of days in the year for each datetime."""

    def __post_init__(self) -> None:
        """Calculate year, julian day and days in year from dates."""
        dateyear = self.dates.astype("datetime64[Y]")
        startnext = (dateyear + 1).astype("datetime64[D]")
        dateday = self.dates.astype("datetime64[D]")
        self.year = dateyear.astype("int") + 1970
        self.julian_day = (dateday - dateyear + 1).astype("int")
        self.days_in_year = (startnext - dateyear).astype("int")
        self.n_dates = len(self.dates)

    def __iter__(self) -> Generator[CalendarDay, Any, Any]:
        """Yield each date in the Calendar in sequence."""
        for idx in range(self.n_dates):
            yield self[idx]

    def __getitem__(self, idx: int) -> CalendarDay:
        """Extract dates by index."""
        return CalendarDay(
            date=self.dates[idx],
            year=self.year[idx],
            julian_day=self.julian_day[idx],
            days_in_year=self.days_in_year[idx],
        )

    def __len__(self) -> int:
        """Length of a Calendar object."""
        return self.n_dates

    def __repr__(self) -> str:
        """Representation of a Calendar instance."""

        return f"Calendar({self.dates[0]!s}, {self.dates[-1]!s})"



"""The :mod:`~pyrealm.constants.core_const` module provides a data class of constants
used in the functions in the :mod:`~pyrealm.core` submodules.
"""  # noqa: D205

from dataclasses import dataclass, field

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants import ConstantsClass


@dataclass(frozen=True)
class CoreConst(ConstantsClass):
    r"""Constants for the :mod:`~pyrealm.core`  module.

    This data class provides constants used in :mod:`~pyrealm.core`, including
    hygrometric conversions, the calculation of water density and viscosity and the
    calculation of atmospheric pressure.

    Sources for some sets of parameters are shown below:

    * **Density of water using Fisher**. Values for the Tumlirz equation taken from
      Table 5 of :cite:t:`Fisher:1975tm`:

      .. math::

            V_p = V_\infty + \dfrac{\lambda}{P_o + P},

      where :math:`\lambda`
      (:attr:`~pyrealm.constants.core_const.CoreConst.fisher_dial_lambda`),
      :math:`P_o` (:attr:`~pyrealm.constants.core_const.CoreConst.fisher_dial_Po`) and
      :math:`V_\infty`
      (:attr:`~pyrealm.constants.core_const.CoreConst.fisher_dial_Vinf`) are all
      temperature dependent polynomial functions.

    * **Density of water using Chen**. Values taken from :cite:t:`chen:2008a` and
      calculated as the inverse of their equation for the specific volume of water
      (:math:`V`):

        .. math::
            V = V^0 - V^0 P/(K^0 +AP+BP^2),

      where :math:`V^0` (:attr:`~pyrealm.constants.core_const.CoreConst.chen_po`),
      :math:`K^0` (:attr:`~pyrealm.constants.core_const.CoreConst.chen_ko`),
      :math:`A` (:attr:`~pyrealm.constants.core_const.CoreConst.chen_ca`, and
      :math:`B` (:attr:`~pyrealm.constants.core_const.CoreConst.chen_cb`) are all
      temperature dependent polynomial functions.

    * **Viscosity of water**. Values for the parameterisation taken from Table 2 and 3
      of :cite:t:`Huber:2009fy`:
      (:attr:`~pyrealm.constants.core_const.CoreConst.huber_tk_ast`,
      :attr:`~pyrealm.constants.core_const.CoreConst.huber_rho_ast`,
      :attr:`~pyrealm.constants.core_const.CoreConst.huber_mu_ast`,
      :attr:`~pyrealm.constants.core_const.CoreConst.huber_H_i`,
      :attr:`~pyrealm.constants.core_const.CoreConst.huber_H_ij`)

    """

    # Universal constants
    k_R: float = 8.3145
    """Universal gas constant (:math:`R` , 8.3145, J/mol/K)"""
    k_co: float = 209476.0
    """O2 partial pressure, Standard Atmosphere (:math:`co` , 209476.0, ppm)"""
    k_c_molmass: float = 12.0107
    """Molecular mass of carbon (:math:`c_molmass` , 12.0107, g)"""
    k_water_molmass: float = 18.01258
    """Molecular mass of water (:math:`h2o_molmass` , 18.01258, g)"""
    k_Po: float = 101325.0
    """Standard reference atmosphere (Allen, 1973) (:math:`P_o` , 101325.0, Pa)"""
    k_To: float = 298.15
    """Standard reference temperature (:math:`T_o` ,  298.15, K)"""
    k_L: float = 0.0065
    """Adiabiatic temperature lapse rate (Allen, 1973)   (:math:`L` , 0.0065, K/m)"""
    k_G: float = 9.80665
    """Gravitational acceleration (:math:`G` , 9.80665, m/s^2)"""
    k_Ma: float = 0.028963
    """Molecular weight of dry air (Tsilingiris, 2008)  (:math:`M_a`, 0.028963,
    kg/mol)"""
    k_Mv: float = 0.01802
    """Molecular weight of water vapour (Tsilingiris, 2008)  (:math:`M_v`,0.01802,
    kg/mol)"""
    k_CtoK: float = 273.15
    """Conversion from °C to K   (:math:`CtoK` , 273.15, -)"""
    k_pir = np.pi / 180.0
    """Conversion factor from radians to degrees   (``pir`` , ~0.01745, -)"""

    # TODO - these might be better in a separate SplashConst
    k_w = 0.26
    """Entrainment factor (Lhomme, 1997; Priestley & Taylor, 1972)"""
    k_Cw = 1.05
    """Supply constant, mm/hr (Federer, 1982)"""

    # Solar constants
    visible_light_albedo: float = 0.03
    """The visible light albedo (:math:`A_{vis}`, unitless, Sellers, 1985)."""

    swdown_to_ppfd_factor: float = 2.04
    """Conversion factor from shortwave downwelling radiation (W m-2) to photosynthetic
    photon flux density (PPFD, µmol m-2 s-1): one W m-2 of sunlight is roughly 4.57 µmol
    m-2 s-1 of full spectrum sunlight, of which about 46% (2.04 / 4.57) is PPFD. (Meek
    et al., 1984)."""

    transmissivity_coef: tuple[float, float, float] = (0.25, 0.5, 2.67e-5)
    """Coefficients for calculating transmissivity from :cite:t:`Linacre:1968a`: cloudy
    transmissivity (:math:`c=0.25`),  angular coefficient of transmittivity
    (:math:`d=0.5`) and elevation factor (:math:`f`=2.67e-5`)."""

    net_longwave_radiation_coef: tuple[float, float] = (0.2, 107.0)
    """Coefficients (:math:`b, A`) of net longwave radiation function, Eqn. 11 and Table
    1 of :cite:t:`colinprentice:1993a`"""

    shortwave_albedo: float = 0.17
    """The shortwave albedo (:math:`A_{sw}`, unitless, Federer, 1968)."""

    solar_constant: float = 1360.8
    """The solar constant (:math:`G_{sc}`, W/m^2, Kopp & Lean, 2011) - the long term
    mean total solar irradiance."""

    day_seconds: float = 86400
    """The number of seconds in one solar day."""

    equation_of_time_coef: tuple[float, ...] = (
        0.000075,
        0.001868,
        -0.032077,
        -0.014615,
        -0.04089,
        229.18,
    )
    """Coefficients of the equation of time :cite:t:`iqbal:1983a`"""

    # Paleoclimate variables:
    solar_eccentricity: float = 0.0167
    """Solar eccentricity (:math:`e`), using default value for 2000 CE 
    :cite:t:`berger:1978a`."""
    solar_obliquity: float = 23.44
    r"""Solar obliquity in degrees (:math:`\epsilon`), using default value for 2000 CE
    :cite:t:`berger:1978a`."""
    solar_perihelion: float = 283.0
    r"""Solar longitude of perihelion in degrees (:math:`\omega`), using default value
    for 2000 CE :cite:t:`berger:1978a`."""

    # Hygro constants
    magnus_coef: NDArray[np.float64] = field(
        default_factory=lambda: np.array((611.2, 17.62, 243.12))
    )
    """Three coefficients of the Magnus equation for saturated vapour pressure,
    defaulting to those of  ``Sonntag1990``."""
    mwr: float = 0.622
    """The ratio molecular weight of water vapour to dry air (:math:`MW_r`, -)"""
    magnus_option: str | None = None
    """Pre-defined Magnus equation parameterisations. Use one of ``Allen1998``,
    ``Alduchov1996`` or ``Sonntag1990`` to set values for
    :attr:`~pyrealm.constants.core_const.CoreConst.magnus_coef`.
    """

    # Water constants
    water_density_method: str = "fisher"
    """Set the method used for calculating water density ('fisher' or 'chen')."""

    # Fisher Dial
    fisher_dial_lambda: NDArray[np.float64] = field(
        default_factory=lambda: np.array(
            [1788.316, 21.55053, -0.4695911, 0.003096363, -7.341182e-06]
        )
    )
    r"""Coefficients of the temperature dependent polynomial for :math:`\lambda`
     in the Tumlirz equation."""

    fisher_dial_Po: NDArray[np.float64] = field(
        default_factory=lambda: np.array(
            [5918.499, 58.05267, -1.1253317, 0.0066123869, -1.4661625e-05]
        )
    )
    """Coefficients of the temperature dependent polynomial for :math:`P_0` in the
    Tumlirz equation."""

    fisher_dial_Vinf: NDArray[np.float64] = field(
        default_factory=lambda: np.array(
            [
                0.6980547,
                -0.0007435626,
                3.704258e-05,
                -6.315724e-07,
                9.829576e-09,
                -1.197269e-10,
                1.005461e-12,
                -5.437898e-15,
                1.69946e-17,
                -2.295063e-20,
            ]
        )
    )
    r"""Coefficients of the temperature dependent polynomial for :math:`V_{\infty}`
    in the Tumlirz equation."""

    # Chen water density
    chen_po: NDArray[np.float64] = field(
        default_factory=lambda: np.array(
            [
                0.99983952,
                6.788260e-5,
                -9.08659e-6,
                1.022130e-7,
                -1.35439e-9,
                1.471150e-11,
                -1.11663e-13,
                5.044070e-16,
                -1.00659e-18,
            ]
        )
    )
    r"""Coefficients of the polynomial relationship of water density with temperature at
    1 atm (:math:`P^0`, kg/m^3) from :cite:t:`chen:2008a`."""

    chen_ko: NDArray[np.float64] = field(
        default_factory=lambda: np.array(
            [19652.17, 148.1830, -2.29995, 0.01281, -4.91564e-5, 1.035530e-7]
        )
    )
    r"""Polynomial relationship of bulk modulus of water with temperature at 1 atm
     (:math:`K^0`, kg/m^3) from :cite:t:`chen:2008a`."""

    chen_ca: NDArray[np.float64] = field(
        default_factory=lambda: np.array(
            [3.26138, 5.223e-4, 1.324e-4, -7.655e-7, 8.584e-10]
        )
    )
    r"""Coefficients of the polynomial temperature dependent coefficient :math:`A` from
     :cite:t:`chen:2008a`."""

    chen_cb: NDArray[np.float64] = field(
        default_factory=lambda: np.array(
            [7.2061e-5, -5.8948e-6, 8.69900e-8, -1.0100e-9, 4.3220e-12]
        )
    )
    r"""Coefficients of the polynomial temperature dependent coefficient :math:`B` from
     :cite:t:`chen:2008a`."""

    # Huber
    simple_viscosity: bool = False
    """Boolean setting for use of simple viscosity calculations"""
    huber_tk_ast: float = 647.096
    """Huber reference temperature (:math:`tk_{ast}`, 647.096, Kelvin)"""
    huber_rho_ast: float = 322.0
    r"""Huber reference density (:math:`\rho_{ast}`, 322.0, kg/m^3)"""
    huber_mu_ast: float = 1e-06
    r"""Huber reference pressure (:math:`\mu_{ast}` 1.0e-6, Pa s)"""

    huber_H_i: NDArray[np.float64] = field(
        default_factory=lambda: np.array([1.67752, 2.20462, 0.6366564, -0.241605])
    )
    """Temperature dependent parameterisation of Hi in Huber."""
    huber_H_ij: NDArray[np.float64] = field(
        default_factory=lambda: np.array(
            [
                [0.520094, 0.0850895, -1.08374, -0.289555, 0.0, 0.0],
                [0.222531, 0.999115, 1.88797, 1.26613, 0.0, 0.120573],
                [-0.281378, -0.906851, -0.772479, -0.489837, -0.25704, 0.0],
                [0.161913, 0.257399, 0.0, 0.0, 0.0, 0.0],
                [-0.0325372, 0.0, 0.0, 0.0698452, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.00872102, 0.0],
                [0.0, 0.0, 0.0, -0.00435673, 0.0, -0.000593264],
            ],
        )
    )
    """Temperature and mass density dependent parameterisation of Hij in Huber."""

    def __post_init__(self) -> None:
        """Populate parameters from init settings.

        This checks the init inputs and populates ``magnus_coef`` from the presets
        if no magnus_coef is specified.

        Returns:
            None
        """
        alts = dict(
            Allen1998=np.array((610.8, 17.27, 237.3)),
            Alduchov1996=np.array((610.94, 17.625, 243.04)),
            Sonntag1990=np.array((611.2, 17.62, 243.12)),
        )

        # Note that object is being used here to update a frozen dataclass

        # Parse other options
        if self.magnus_option is not None:
            if self.magnus_option not in alts:
                raise (ValueError(f"magnus_option must be one of {list(alts.keys())}"))

            object.__setattr__(self, "magnus_coef", alts[self.magnus_option])
            return

        if self.magnus_coef is not None and len(self.magnus_coef) != 3:
            raise TypeError("magnus_coef must be a tuple of 3 numbers")



"""The :mod:`~pyrealm.core.pressure` submodule contains core functions for calculating
atmospheric pressure.
"""  # noqa D210, D415

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants import CoreConst


def calc_patm(
    elv: NDArray[np.float64], core_const: CoreConst = CoreConst()
) -> NDArray[np.float64]:
    r"""Calculate atmospheric pressure from elevation.

    Calculates atmospheric pressure as a function of elevation with reference to the
    standard atmosphere.  The elevation-dependence of atmospheric pressure is computed
    by assuming a linear decrease in temperature with elevation and a mean adiabatic
    lapse rate (Eqn 3, :cite:alp:`BerberanSantos:2009bk`):

    .. math::

        p(z) = p_0 ( 1 - L z / K_0) ^{ G M / (R L) },

    Args:
        elv: Elevation above sea-level (:math:`z`, metres above sea level.)
        core_const: Instance of :class:`~pyrealm.constants.core_const.CoreConst`.

    Returns:
        A numeric value for :math:`p` in Pascals.

    Examples:
        >>> # Standard atmospheric pressure, in Pa, corrected for 1000 m.a.s.l.
        >>> round(calc_patm(1000), 2)
        90241.54
    """

    # Convert elevation to pressure, Pa.
    return core_const.k_Po * (1.0 - core_const.k_L * elv / core_const.k_To) ** (
        core_const.k_G * core_const.k_Ma / (core_const.k_R * core_const.k_L)
    )



"""The flora module implements definitions of:

* The ``PlantFunctionalType`` and ``PlantFunctionalTypeStrict`` dataclasses, which are
  used to parameterise the traits of different plant functional types. The
  ``PlantFunctionalType`` dataclass is a subclass of ``PlantFunctionalTypeStrict`` that
  simply adds default values to the attributes.
* The ``PlantFunctionalTypeStrict`` dataclass is used as the basis of a
  :mod:`~marshmallow` schema for validating the creation of plant functional types from
  data files. This intentionally enforces a complete description of the traits in the
  input data. The ``PlantFunctionalType`` is provided as a more convenient API for
  programmatic use.
* The Flora dataclass, which represents a collection of plant functional types for use
  in describing a plant community in a simulation. It provides the same trait attributes
  as the plant functional type classes, but the values are arrays of trait values across
  the provided PFTS. The Flora class also defines factory methods to create instances
  from plant functional type data stored in JSON, TOML or CSV formats.
* The StemTraits dataclass, which represents a collection of stems used in a simulation.
  It again provides the same trait attributes as the plant functional type classes, but
  as arrays. This differs from the Flora class in allowing multiple stems of the same
  plant functional type and is primarily used to broadcast PFT traits into arrays for
  use in calculating demography across the stems in plant cohorts.
"""  # noqa: D415

from __future__ import annotations

import json
import sys
from collections import Counter
from collections.abc import Sequence
from dataclasses import InitVar, dataclass, field, fields
from pathlib import Path
from typing import ClassVar

import marshmallow_dataclass
import numpy as np
import pandas as pd
from marshmallow.exceptions import ValidationError
from numpy.typing import NDArray

from pyrealm.demography.core import (
    CohortMethods,
    PandasExporter,
    _validate_demography_array_arguments,
)

if sys.version_info[:2] >= (3, 11):
    import tomllib
    from tomllib import TOMLDecodeError
else:
    import tomli as tomllib
    from tomli import TOMLDecodeError


def calculate_crown_q_m(
    m: float | NDArray[np.float64], n: float | NDArray[np.float64]
) -> float | NDArray[np.float64]:
    """Calculate the crown scaling trait ``q_m``.

    The value of q_m is a constant crown scaling parameter derived from the ``m`` and
    ``n`` attributes defined for a plant functional type.

    Args:
        m: Crown shape parameter
        n: Crown shape parameter
    """
    return (
        m
        * n
        * ((n - 1) / (m * n - 1)) ** (1 - 1 / n)
        * (((m - 1) * n) / (m * n - 1)) ** (m - 1)
    )


def calculate_crown_z_max_proportion(
    m: float | NDArray[np.float64], n: float | NDArray[np.float64]
) -> float | NDArray[np.float64]:
    r"""Calculate the z_m trait.

    The z_m proportion (:math:`p_{zm}`) is the constant proportion of stem height at
    which the maximum crown radius is found for a given plant functional type.

    .. math::

        p_{zm} = \left(\dfrac{n-1}{m n -1}\right)^ {\tfrac{1}{n}}

    Args:
        m: Crown shape parameter
        n: Crown shape parameter
    """

    return ((n - 1) / (m * n - 1)) ** (1 / n)


@dataclass(frozen=True)
class PlantFunctionalTypeStrict:
    """The PlantFunctionalTypeStrict dataclass.

    This dataclass implements the set of traits required to define a plant functional
    type for use in ``pyrealm``.

    * Most traits are taken from the definition of the T Model of plant growth and GPP
      allocation :cite:`Li:2014bc`.
    * The foliage maintenance respiration fraction was not explicitly included in
      :cite:t:`Li:2014bc` - there was assumed to be a 10% penalty on GPP before
      calculating the other component - but has been explicitly included here.
    * This implementation adds two further crown shape parameters (``m`` and ``n`` and
      ``f_g``). The first two are then used to calculate two constant derived attributes
      (``q_m`` and ``z_max_ratio``) that define the vertical distribution of the crown.
      The last parameter (``f_g``) is the crown gap fraction, that defines the vertical
      distribution of leaves within the crown. This crown model parameterisation
      follows the implementation developed in the PlantFATE model :cite:`joshi:2022a`.

    See also :class:`~pyrealm.demography.flora.PlantFunctionalType` for the default
    values implemented in that subclass.
    """

    name: str
    r"""The name of the plant functional type."""
    a_hd: float
    r"""Initial slope of height-diameter relationship (:math:`a`, -)"""
    ca_ratio: float
    r"""Initial ratio of crown area to stem cross-sectional area
    (:math:`c`, -)"""
    h_max: float
    r"""Maximum tree height (:math:`H_m`, m)"""
    rho_s: float
    r"""Sapwood density (:math:`\rho_s`, kg Cm-3)"""
    lai: float
    """Leaf area index within the crown (:math:`L`,  -)"""
    sla: float
    r"""Specific leaf area (:math:`\sigma`,  m2 kg-1 C)"""
    tau_f: float
    r"""Foliage turnover time (:math:`\tau_f`,years)"""
    tau_rt: float
    r"""Reproductive tissue turnover time (:math:`\tau_rt`,years)"""
    tau_r: float
    r"""Fine-root turnover time (:math:`\tau_r`,  years)"""
    par_ext: float
    r"""Extinction coefficient of photosynthetically active radiation (PAR) (:math:`k`,
     -)"""
    yld: float
    r"""Yield factor (:math:`y`,  -)"""
    zeta: float
    r"""Ratio of fine-root mass to foliage area (:math:`\zeta`, kg C m-2)"""
    resp_r: float
    r"""Fine-root specific respiration rate (:math:`r_r`, year-1)"""
    resp_rt: float
    r"""Reproductive tissue respiration rate (:math:`r_{rt}`, year-1)"""
    resp_s: float
    r"""Sapwood-specific respiration rate (:math:`r_s`,  year-1)"""
    resp_f: float
    r"""Foliage maintenance respiration fraction (:math:`r_f`,  -)"""
    m: float
    r"""Crown shape parameter (:math:`m`, -)"""
    n: float
    r"""Crown shape parameter (:math:`n`, -)"""
    f_g: float
    r"""Crown gap fraction (:math:`f_g`, -)"""

    q_m: float = field(init=False)
    """Scaling factor to derive maximum crown radius from crown area."""
    z_max_prop: float = field(init=False)
    """Proportion of stem height at which maximum crown radius is found."""

    p_foliage_for_reproductive_tissue: float
    """Mass of reproductive tissue as a proportion of foliage mass (:math:`p_{rt}`, -).
    """
    gpp_topslice: float
    """Proportion of GPP to topslice before allocation."""

    def __post_init__(self) -> None:
        """Populate derived attributes.

        This method populates the ``q_m`` and ``z_max_ratio`` attributes from the
        provided values of ``m`` and ``n``.
        """

        # Calculate q_m and z_max proportion. Need to use __setattr__ because the
        # dataclass is frozen.
        object.__setattr__(self, "q_m", calculate_crown_q_m(m=self.m, n=self.n))
        object.__setattr__(
            self, "z_max_prop", calculate_crown_z_max_proportion(m=self.m, n=self.n)
        )


@dataclass(frozen=True)
class PlantFunctionalType(PlantFunctionalTypeStrict):
    r"""The PlantFunctionalType dataclass.

    This dataclass is a subclass of
    :class:`~pyrealm.demography.flora.PlantFunctionalTypeStrict` that implements exactly
    the same set of traits but provides default values. This class is intended as a
    convenience API for programmatic use, where the parent provides a strict schema for
    generating plant functional type instances from data.

    The table below lists the attributes and default values taken from Table 1 of
    :cite:t:`Li:2014bc`, except for ``m``, ``n`` and ``f_g`` which take representative
    values from :cite:t:`joshi:2022a`.

    .. csv-table::
        :header: "Attribute", "Default", "Unit"
        :widths: 15, 10, 30

        a_hd, 116.0, -
        ca_ratio, 390.43, -
        h_max,  25.33, m
        rho_s, 200.0, kg Cm-3
        lai,  1.8, -
        sla,  14.0, m2 kg-1 C
        tau_f, 4.0, years
        tau_r,  1.04, years
        par_ext, 0.5, -
        yld, 0.6, -
        zeta, 0.17, kg C m-2
        resp_r,  0.913, year-1
        resp_s,  0.044, year-1
        resp_f,  0.1, -
        m, 2, -
        n, 5, -
        f_g, 0.05, -
    """

    name: str
    a_hd: float = 116.0
    ca_ratio: float = 390.43
    h_max: float = 25.33
    rho_s: float = 200.0
    lai: float = 1.8
    sla: float = 14.0
    tau_f: float = 4.0
    tau_rt: float = 1.0  # Default value 1 as this is a denominator
    tau_r: float = 1.04
    par_ext: float = 0.5
    yld: float = 0.6
    zeta: float = 0.17
    resp_r: float = 0.913
    resp_s: float = 0.044
    resp_f: float = 0.1
    resp_rt: float = 0.0
    m: float = 2
    n: float = 5
    f_g: float = 0.05
    p_foliage_for_reproductive_tissue: float = 0.0
    gpp_topslice: float = 0.0


PlantFunctionalTypeSchema = marshmallow_dataclass.class_schema(
    PlantFunctionalTypeStrict
)
"""Marshmallow validation schema class for validating PlantFunctionalType data.

This schema explicitly uses the strict version of the dataclass, which enforces complete
descriptions of plant functional type data rather than allowing partial data and filling
in gaps from the default values.
"""


@dataclass(frozen=True)
class Flora(PandasExporter):
    """A dataclass providing trait data on collection of plant functional types.

    A flora provides trait data on the complete collection of plant functional types
    that will be used within a particular simulation. The dataclass provides access
    to trait attributes as row arrays across those plant functional types.

    The class is created using a list of
    :class:`~pyrealm.demography.flora.PlantFunctionalType` or
    :class:`~pyrealm.demography.flora.PlantFunctionalTypeStrict` instances, which must
    have unique names.

    Args:
        pfts: A sequence of ``PlantFunctionalType`` or ``PlantFunctionalTypeStrict``
            instances, which must not have duplicated
            :attr:`~pyrealm.demography.flora.PlantFunctionalTypeStrict.name` attributes.
    """

    # The only init argument.
    pfts: InitVar[Sequence[type[PlantFunctionalTypeStrict]]]
    r"""A sequence of plant functional type instances to include in the Flora."""

    # A class variable setting the names of PFT traits held as arrays.
    array_attrs: ClassVar[tuple[str, ...]] = tuple(
        f.name for f in fields(PlantFunctionalTypeStrict)
    )

    # Populated post init
    # - trait arrays
    name: NDArray[np.str_] = field(init=False)
    r"""The name of the plant functional type."""
    a_hd: NDArray[np.float64] = field(init=False)
    r"""Initial slope of height-diameter relationship (:math:`a`, -)"""
    ca_ratio: NDArray[np.float64] = field(init=False)
    r"""Initial ratio of crown area to stem cross-sectional area
    (:math:`c`, -)"""
    h_max: NDArray[np.float64] = field(init=False)
    r"""Maximum tree height (:math:`H_m`, m)"""
    rho_s: NDArray[np.float64] = field(init=False)
    r"""Sapwood density (:math:`\rho_s`, kg Cm-3)"""
    lai: NDArray[np.float64] = field(init=False)
    """Leaf area index within the crown (:math:`L`,  -)"""
    sla: NDArray[np.float64] = field(init=False)
    r"""Specific leaf area (:math:`\sigma`,  m2 kg-1 C)"""
    tau_f: NDArray[np.float64] = field(init=False)
    r"""Foliage turnover time (:math:`\tau_f`,years)"""
    tau_r: NDArray[np.float64] = field(init=False)
    r"""Fine-root turnover time (:math:`\tau_r`,  years)"""
    tau_rt: NDArray[np.float64] = field(init=False)
    r"""Reproductive tissue turnover time (:math:`\tau_rt`,years)"""
    par_ext: NDArray[np.float64] = field(init=False)
    r"""Extinction coefficient of photosynthetically active radiation (PAR) (:math:`k`,
     -)"""
    yld: NDArray[np.float64] = field(init=False)
    r"""Yield factor (:math:`y`,  -)"""
    zeta: NDArray[np.float64] = field(init=False)
    r"""Ratio of fine-root mass to foliage area (:math:`\zeta`, kg C m-2)"""
    resp_r: NDArray[np.float64] = field(init=False)
    r"""Fine-root specific respiration rate (:math:`r_r`, year-1)"""
    resp_s: NDArray[np.float64] = field(init=False)
    r"""Sapwood-specific respiration rate (:math:`r_s`,  year-1)"""
    resp_f: NDArray[np.float64] = field(init=False)
    r"""Foliage maintenance respiration fraction (:math:`r_f`,  -)"""
    resp_rt: NDArray[np.float64] = field(init=False)
    r"""Reproductive tissue respiration rate (:math:`r_{rt}`,  -)"""
    m: NDArray[np.float64] = field(init=False)
    r"""Crown shape parameter (:math:`m`, -)"""
    n: NDArray[np.float64] = field(init=False)
    r"""Crown shape parameter (:math:`n`, -)"""
    f_g: NDArray[np.float64] = field(init=False)
    r"""Crown gap fraction (:math:`f_g`, -)"""
    q_m: NDArray[np.float64] = field(init=False)
    """Scaling factor to derive maximum crown radius from crown area."""
    z_max_prop: NDArray[np.float64] = field(init=False)
    """Proportion of stem height at which maximum crown radius is found."""

    # - other instance attributes
    pft_dict: dict[str, type[PlantFunctionalTypeStrict]] = field(init=False)
    """A dictionary of the original plant functional type instances, keyed by name."""
    pft_indices: dict[str, int] = field(init=False)
    """An dictionary giving the index of each PFT name in the trait array attributes."""
    n_pfts: int = field(init=False)
    """The number of plant functional types in the Flora instance."""
    _n_stems: int = field(init=False)
    """Private attribute for compatibility with StemTraits API."""

    p_foliage_for_reproductive_tissue: NDArray[np.float64] = field(init=False)
    """Proportion of foliage used to calculate reproductive tissue."""
    gpp_topslice: NDArray[np.float64] = field(init=False)
    """Proportion of GPP to topslice before allocation."""

    def __post_init__(self, pfts: Sequence[type[PlantFunctionalTypeStrict]]) -> None:
        # Check the PFT data
        if (not isinstance(pfts, Sequence)) or (
            not all([isinstance(v, PlantFunctionalTypeStrict) for v in pfts])
        ):
            raise ValueError(
                "The pfts argument must be a sequence of PlantFunctionalType instances"
            )

        # Validate the PFT instances - check there are no duplicate PFT names.
        pft_names = Counter([p.name for p in pfts])
        duplicates = [k for k, v in pft_names.items() if v > 1]

        if duplicates:
            raise ValueError(
                f"Duplicated plant functional type names: {','.join(duplicates)}"
            )

        # Populate the pft dictionary using the PFT name as key and the number of PFTs
        object.__setattr__(self, "pft_dict", {p.name: p for p in pfts})
        object.__setattr__(self, "n_pfts", len(pfts))
        object.__setattr__(self, "_n_stems", len(pfts))

        # Populate the trait attributes with arrays
        for pft_field in self.array_attrs:
            object.__setattr__(
                self, pft_field, np.array([getattr(pft, pft_field) for pft in pfts])
            )

        # Populate the pft trait indices
        object.__setattr__(self, "pft_indices", {v: k for k, v in enumerate(self.name)})

    def __repr__(self) -> str:
        """Simple representation of the Flora instance."""

        return f"Flora with {self._n_stems} functional types: {', '.join(self.name)}"

    @classmethod
    def _from_file_data(cls, file_data: dict) -> Flora:
        """Create a Flora object from a JSON string.

        Args:
            file_data: The payload from a data file defining plant functional types.
        """
        try:
            pfts = PlantFunctionalTypeSchema().load(file_data["pft"], many=True)  # type: ignore[attr-defined]
        except ValidationError as excep:
            raise excep

        return cls(pfts=pfts)

    @classmethod
    def from_json(cls, path: Path) -> Flora:
        """Create a Flora object from a JSON file.

        Args:
            path: A path to a JSON file of plant functional type definitions.
        """

        try:
            file_data = json.load(open(path))
        except (FileNotFoundError, json.JSONDecodeError) as excep:
            raise excep

        return cls._from_file_data(file_data=file_data)

    @classmethod
    def from_toml(cls, path: Path) -> Flora:
        """Create a Flora object from a TOML file.

        Args:
            path: A path to a TOML file of plant functional type definitions.
        """

        try:
            file_data = tomllib.load(open(path, "rb"))
        except (FileNotFoundError, TOMLDecodeError) as excep:
            raise excep

        return cls._from_file_data(file_data)

    @classmethod
    def from_csv(cls, path: Path) -> Flora:
        """Create a Flora object from a CSV file.

        Args:
            path: A path to a CSV file of plant functional type definitions.
        """

        try:
            data = pd.read_csv(path)
        except (FileNotFoundError, pd.errors.ParserError) as excep:
            raise excep

        return cls._from_file_data({"pft": data.to_dict(orient="records")})

    def get_stem_traits(self, pft_names: NDArray[np.str_]) -> StemTraits:
        """Generates a stem traits object for a set of names.

        Args:
            pft_names: An array of PFT names for each stem.
        """

        # Check the initial PFT values are known
        unknown_pfts = set(pft_names).difference(self.name)

        if unknown_pfts:
            raise ValueError(
                f"Plant functional types unknown in flora: {','.join(unknown_pfts)}"
            )

        # Get the indices for the cohort PFT names in the flora PFT data
        pft_index = [self.pft_indices[str(nm)] for nm in pft_names]

        # For each trait, get that attribute from the Flora, extract the values matching
        # the pft_names and pass that into the StemTraits constructor. Validation is
        # turned off here, because the creation method guarantees the data is properly
        # formatted.
        return StemTraits(
            **{trt: getattr(self, trt)[pft_index] for trt in self.array_attrs},
            validate=False,
        )


@dataclass()
class StemTraits(PandasExporter, CohortMethods):
    """A dataclass for stem traits.

    This dataclass is used to provide arrays of plant functional type (PFT) traits
    across a set of stems. The main use case is to provide stem trait data as arrays
    across the cohorts within a community object.

    It provides the same attribute interface as the
    :class:`~pyrealm.demography.flora.Flora` class, but unlike that class:

    * is purely a data container, and
    * plant functional types can be represented multiple times to represent multiple
      stems or cohorts of the same PFT.
    """

    # A class variable setting the attribute names of traits.
    array_attrs: ClassVar[tuple[str, ...]] = tuple(
        f.name for f in fields(PlantFunctionalTypeStrict)
    )
    count_attr: ClassVar[str] = "_n_stems"

    # Instance trait attributes
    name: NDArray[np.str_]
    r"""The name of the plant functional type."""
    a_hd: NDArray[np.float64]
    r"""Initial slope of height-diameter relationship (:math:`a`, -)"""
    ca_ratio: NDArray[np.float64]
    r"""Initial ratio of crown area to stem cross-sectional area
    (:math:`c`, -)"""
    h_max: NDArray[np.float64]
    r"""Maximum tree height (:math:`H_m`, m)"""
    rho_s: NDArray[np.float64]
    r"""Sapwood density (:math:`\rho_s`, kg Cm-3)"""
    lai: NDArray[np.float64]
    """Leaf area index within the crown (:math:`L`,  -)"""
    sla: NDArray[np.float64]
    r"""Specific leaf area (:math:`\sigma`,  m2 kg-1 C)"""
    tau_f: NDArray[np.float64]
    r"""Foliage turnover time (:math:`\tau_f`,years)"""
    tau_rt: NDArray[np.float64]
    r"""Reproductive tissue turnover time (:math:`\tau_rt`,years)"""
    tau_r: NDArray[np.float64]
    r"""Fine-root turnover time (:math:`\tau_r`,  years)"""
    par_ext: NDArray[np.float64]
    r"""Extinction coefficient of photosynthetically active radiation (PAR) (:math:`k`,
     -)"""
    yld: NDArray[np.float64]
    r"""Yield factor (:math:`y`,  -)"""
    zeta: NDArray[np.float64]
    r"""Ratio of fine-root mass to foliage area (:math:`\zeta`, kg C m-2)"""
    resp_r: NDArray[np.float64]
    r"""Fine-root specific respiration rate (:math:`r_r`, year-1)"""
    resp_s: NDArray[np.float64]
    r"""Sapwood-specific respiration rate (:math:`r_s`,  year-1)"""
    resp_f: NDArray[np.float64]
    r"""Foliage maintenance respiration fraction (:math:`r_f`,  -)"""
    resp_rt: NDArray[np.float64]
    r"""Reproductive tissue respiration rate (:math:`r_{rt}`,  -)"""
    m: NDArray[np.float64]
    r"""Crown shape parameter (:math:`m`, -)"""
    n: NDArray[np.float64]
    r"""Crown shape parameter (:math:`n`, -)"""
    f_g: NDArray[np.float64]
    r"""Crown gap fraction (:math:`f_g`, -)"""
    q_m: NDArray[np.float64]
    """Scaling factor to derive maximum crown radius from crown area."""
    z_max_prop: NDArray[np.float64]
    """Proportion of stem height at which maximum crown radius is found."""

    p_foliage_for_reproductive_tissue: NDArray[np.float64]
    """Proportion of foliage used to calculate reproductive tissue."""
    gpp_topslice: NDArray[np.float64]
    """Proportion of GPP to topslice before allocation."""

    validate: bool = True
    """Boolean flag to control validation of the input array sizes."""

    # Post init attributes
    _n_stems: int = field(init=False)

    def __post_init__(self) -> None:
        """Post init validation and attribute setting."""

        if self.validate:
            _validate_demography_array_arguments(
                trait_args={k: getattr(self, k) for k in self.array_attrs}
            )

        self._n_stems = len(self.a_hd)



"""Class to compute the fAPAR_max and annual peak Leaf Area Index (LAI)."""

import numpy as np
from numpy.typing import NDArray
from typing_extensions import Self

from pyrealm.constants import PhenologyConst
from pyrealm.core.experimental import warn_experimental
from pyrealm.core.time_series import AnnualValueCalculator
from pyrealm.core.utilities import check_input_shapes
from pyrealm.pmodel import AcclimationModel, PModel


def check_datetimes(datetimes: NDArray[np.datetime64]) -> None:
    """Check that the datetimes are in a valid format."""

    deltas = datetimes[1:] - datetimes[:-1]
    unique_deltas = np.unique(deltas)

    # check that we have uniformly sampled data
    if np.size(unique_deltas) > 1:
        raise ValueError("datetimes are not evenly spaced.")

    dates = datetimes.astype("datetime64[D]")
    unique_dates, date_counts = np.unique(dates, return_counts=True)
    if not date_counts.max() == date_counts.min():
        raise ValueError("Differing date counts per day")

    obs_per_date = date_counts.max()

    # Data needs to start in northern or southern hemisphere midwinter
    first_month = unique_dates[0].astype("datetime64[M]").astype(str)
    if not (first_month.endswith("01") | first_month.endswith("07")):
        raise ValueError("Data does not start in January or July.")

    ## This does not work for fortnightly data
    # no_leapdays = leapdays(
    #    int(str(datetimes[0].astype("datetime64[Y]"))),
    #    int(str(datetimes[-1].astype("datetime64[Y]"))),
    # )
    #
    # year_remainder = len(unique_dates) % 365
    # check that we have the right number of leap days
    #    if year_remainder > no_leapdays:
    #        raise ValueError("Datetimes do not cover full years.")

    if obs_per_date > 1:
        # subdaily

        # Check that the number of seconds in a day is evenly divisible by the
        # number of observations per day (should already have led to differing date
        # counts).
        day_remainder = (24 * 60 * 60) % obs_per_date
        if day_remainder:
            raise ValueError("Datetime spacing is not evenly divisible into a day.")


def get_annual(
    x: NDArray,
    datetimes: NDArray[np.datetime64],
    growing_season: NDArray[np.bool],
    method: str,
) -> NDArray:
    """Computes an array of the annual total or mean of an entity x given datetimes.

    Args:
        x: Array of values to be converted to annual values. Should be either daily (
            same datetimes as growing_season) or subdaily (same datetimes as datetimes
            array)
        datetimes: Datetimes of the measurements as np.datetime64 arrays.
        growing_season: Bool array of days, indicating whether they are ain growing
            season or not.
        method: Either "total" (sum all values of the year) or "mean" (take the mean
            of all values of the year)
    """

    # Extract years from datetimes
    all_years = datetimes.astype("datetime64[Y]")

    if len(x) == len(growing_season):  # this is daily data
        daily_x = x
        n_days = len(x)
        years_by_day = all_years.view()
        obs_per_day = int(len(datetimes) / n_days)
        years_by_day.shape = tuple([n_days, obs_per_day, *list(all_years.shape[1:])])
    elif len(x) == len(datetimes):  # this is subdaily data
        # Create scaler object to handle conversion between scales
        scaler = AcclimationModel(datetimes)
        scaler.set_nearest(np.timedelta64(12, "h"))
        # Convert values to daily to match with growing_season
        daily_x = scaler.get_daily_means(x)
        years_by_day = scaler.get_window_values(np.asarray(all_years))
    else:
        raise ValueError("Input array does not fit datetimes nor growing_season array")

    # Which years are present?
    years = np.unique(all_years)

    # Compute annual totals or means, only taking into account the days which are in
    # growing season.

    annual_x = np.zeros(len(years))

    if method == "total":
        for i in range(len(years)):
            annual_x[i] = np.sum(
                daily_x[growing_season & (years_by_day[:, 0] == years[i])]
            )
    elif method == "mean":
        for i in range(len(years)):
            annual_x[i] = np.mean(
                daily_x[growing_season & (years_by_day[:, 0] == years[i])]
            )
    else:
        raise ValueError("No valid method given for annual values")

    return annual_x


def daily_to_subdaily(
    x: NDArray,
    datetimes: NDArray[np.datetime64],
) -> NDArray:
    """Broadcasts an array of the entity x from daily values to subdaily values.

    Args:
        x: Array of daily values.
        datetimes: Subdaily datetimes as np.datetime64 array.
    """

    n_days = len(x)
    obs_per_day = int(len(datetimes) / n_days)

    subdaily_x = np.repeat(x, obs_per_day)

    return subdaily_x


class FaparLimitation:
    r"""FaparLimitation class to compute fAPAR_max and LAI.

    This class takes the annual total potential GPP and precipitation, the annual mean
    CA, Chi and VPD during the growing season, as well as the aridity index and some
    constants to compute the annual peak fractional absorbed photosynthetically active
    radiation (fAPAR_max) and annual peak Leaf Area Index (LAI).

    .. todo::

        * Allow for other timescales than daily or subdaily (e.g. monthly)
        * Daily conditions are taken as noon values - might need to relax this.
        * Growing season definition -  users currently need to provide their own growing
          season definition in `from_pmodel`
    """

    __experimental__ = True

    def _check_shapes(self) -> None:
        """Internal class to check all the input arrays have the same size."""

        check_input_shapes(
            self.annual_total_potential_gpp,
            self.annual_mean_ca,
            self.annual_mean_chi,
            self.annual_mean_vpd,
            self.annual_total_precip,
            self.aridity_index,
        )

    def __init__(
        self,
        annual_total_potential_gpp: NDArray[np.float64],
        annual_mean_ca: NDArray[np.float64],
        annual_mean_chi: NDArray[np.float64],
        annual_mean_vpd: NDArray[np.float64],
        annual_total_precip: NDArray[np.float64],
        aridity_index: NDArray[np.float64],
    ) -> None:
        r"""Annual peak fractional absorbed photosynthetically active radiation (fAPAR).

        Computes fAPAR_max as minimum of a water-limited and an energy-limited
        quantity.

        Args:
            annual_total_potential_gpp: Aka A_0, the annual sum of potential GPP.
              Potential GPP would be achieved if fAPAR = 1. [mol m^{-2} year^{-1}]
            annual_mean_ca: Ambient CO2 partial pressure. [Pa]
            annual_mean_chi: Annual mean ratio of leaf-internal CO2 partial pressure to
               c_a during the growing season (>0℃). [Pa]
            annual_mean_vpd: Aka D, annual mean vapour pressure deficit (VPD) during the
              growing season (>0℃). [Pa]
            annual_total_precip: Aka P, annual total precipitation. [mol m^{-2}
             year^{-1}]
            aridity_index: Aka AI, climatological estimate of local aridity index.
        """

        warn_experimental("FaparLimitation")

        self.annual_total_potential_gpp = annual_total_potential_gpp
        """Aka A_0, the annual sum of potential GPP. [mol m^{-2} year^{-1}]"""
        self.annual_mean_ca = annual_mean_ca
        """Ambient CO2 partial pressure. [Pa]"""
        self.annual_mean_chi = annual_mean_chi
        """Annual mean ratio of leaf-internal CO2 partial pressure to c_a during the 
        growing season (>0℃). [Pa]"""
        self.annual_mean_vpd = annual_mean_vpd
        """Annual mean vapour pressure deficit (VPD) during the growing season [Pa]"""
        self.annual_total_precip = annual_total_precip
        """Annual total precipitation. [mol m^{-2} year^{-1}]"""
        self.aridity_index = aridity_index
        """Climatological estimate of local aridity index."""

        self._check_shapes()

        # Make sure the aridity index is not zero
        if aridity_index <= 0:
            raise ValueError("The aridity index has to be positive.")

        # Constants used for phenology computations
        self.phenology_const = PhenologyConst()

        #  f_0 is the ratio of annual total transpiration of annual total
        #  precipitation, which is an empirical function of the climatic Aridity Index
        #  (AI).
        a, b, c = self.phenology_const.f0_coefficients
        f_0 = a * np.exp(-b * np.log(aridity_index / c) ** 2)

        fapar_energylim = 1.0 - self.phenology_const.z / (
            self.phenology_const.k * annual_total_potential_gpp
        )
        fapar_waterlim = (
            f_0
            * annual_total_precip
            * annual_mean_ca
            * (1 - annual_mean_chi)
            / (1.6 * annual_mean_vpd * annual_total_potential_gpp)
        )

        self.fapar_max = np.minimum(fapar_waterlim, fapar_energylim)
        """Maximum fapar given water or energy limitation for each year."""
        self.energy_limited = fapar_energylim < fapar_waterlim
        """Is fapar_max limited by water or energy for each year."""
        self.annual_precip_molar = annual_total_precip
        """The annual precipitation in moles for each year."""

        self.lai_max = -(1 / self.phenology_const.k) * np.log(1.0 - self.fapar_max)

    @classmethod
    def from_pmodel(
        cls,
        pmodel: PModel,
        growing_season: NDArray[np.bool],
        datetimes: NDArray[np.datetime64],
        precip: NDArray[np.float64],
        aridity_index: NDArray[np.float64],
        gpp_penalty_factor: NDArray[np.float64] | None = None,
    ) -> Self:
        r"""Get FaparLimitation from PModel input.

        Computes the input for fAPAR_max from the P Model and additional inputs.

        Args:
            pmodel: pyrealm.pmodel.PModel
            growing_season: Bool array indicating which times are within growing
              season by some definition and implementation.
            datetimes: Array of datetimes to consider.
            precip: Precipitation for given datetimes.
            aridity_index: Climatological estimate of local aridity index.
            gpp_penalty_factor: Penalty factor to be applied to pmodel.gpp
        """

        check_datetimes(datetimes)

        avc = AnnualValueCalculator(
            timing=datetimes,
            growing_season=growing_season,
        )

        # Get the total GPP for each observation
        # - need to convert P Model GPP in µg C m-2 s-1 into moles C m-2 y-1
        # - also need to handle missing values, easier to take _mean_ annual value
        #   and scale it up to an annual total
        # - TODO - handle incompleteness - when do we stop estimating annual values from
        #   partial years (or at least warn about it)

        # Extract GPP and apply any observation level penalty factor
        total_gpp = pmodel.gpp
        if gpp_penalty_factor is not None:
            total_gpp *= gpp_penalty_factor

        # Calculate annual mean potential GPP and scale up to the year
        annual_mean_potential_gpp = avc.get_annual_means(total_gpp)
        annual_total_potential_gpp = (
            annual_mean_potential_gpp * (avc.year_n_days) * 86400 * 1e-6
        ) / pmodel.core_const.k_c_molmass

        # Calculate annual mean ca, chi and VPD within growing season
        annual_mean_ca = avc.get_annual_means(pmodel.env.ca, within_growing_season=True)
        annual_mean_chi = avc.get_annual_means(
            pmodel.optchi.chi, within_growing_season=True
        )
        annual_mean_vpd = avc.get_annual_means(
            pmodel.env.vpd, within_growing_season=True
        )

        # Calculate total annual precipitation
        annual_total_precip = avc.get_annual_totals(precip)

        return cls(
            annual_total_potential_gpp=annual_total_potential_gpp,
            annual_mean_ca=annual_mean_ca,
            annual_mean_chi=annual_mean_chi,
            annual_mean_vpd=annual_mean_vpd,
            annual_total_precip=annual_total_precip,
            aridity_index=aridity_index,
        )



"""This module implements a version of the two leaf, two stream model of assimilation
:cite:p:`depury:1997a`. The implementation is intended to align closely with the similar
implementation in the BESS model.

The module provides two core classes:

* The :class:`~pyrealm.pmodel.two_leaf.TwoLeafIrradiance` class can be used to estimate
  the irradiance absorbed by sunlit and shaded leaves, given the solar elevation angle,
  the atmospheric pressure, the leaf area index and the photosynthetic photon flux
  density.

* The :class:`~pyrealm.pmodel.two_leaf.TwoLeafAssimilation` then takes an instance of
  the irradiance class and a fitted  :class:`~pyrealm.pmodel.pmodel.PModel` or
  :class:`~pyrealm.pmodel.pmodel.SubdailyPModel`. The carboxylation rate and optimal chi
  calculations from the P Model are then used to calculate the total assimilation by
  sunlit and shaded leaves.

The irradiance calculations are independent of the subsequent calculation of
assimilation, so these classes are separated to allow the same irradiance estimates to
be reused with different assimilation models.
"""  # noqa D210, D415

from warnings import warn

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants.core_const import CoreConst
from pyrealm.constants.two_leaf import TwoLeafConst
from pyrealm.core.bounds import BoundsChecker
from pyrealm.core.experimental import warn_experimental
from pyrealm.core.utilities import check_input_shapes
from pyrealm.pmodel.pmodel import PModel, SubdailyPModel

# ------------------------------
# Irradiance class and functions
# ------------------------------


class TwoLeafIrradiance:
    r"""Calculate the irradiance absorbed by sunlit and shaded leaves.

    The two leaf, two stream model of :cite:t:`depury:1997a` partitions the irradiance
    absorbed by the canopy into the irradiance absorbed by sunlit (:math:`I_{csun}`) and
    shaded (:math:`I_{cshade}`) leaves.

    These irradiances are calculated using the canopy leaf area index (:math:`L`) and
    then the solar elevation (:math:`\beta`), photosynthetic photon flux density (PPFD)
    and atmospheric pressure (:math:`P`) as follows:

    * The fraction of diffuse light (:math:`f_d`) is calculated (see
      :meth:`calculate_fraction_of_diffuse_radiation`) and used to partition the
      incoming PPFD into the beam (:math:`I_b = \textrm{PPFD}(1-f_d)`) and diffuse
      (:math:`I_d=PPFD f_d`) irradiances reaching the canopy.

    * Extinction coefficients are calculated for both beam (:math:`k_b`) and scattered
      light (:math:`k_b'`) reaching the canopy, given the solar elevation :math:`\beta`
      (see :meth:`calculate_beam_extinction_coef`).

    * Canopy reflectance coefficients are calculated for both beam and diffuse light.
      The beam reflectance (:math:`\rho_{cb}`) varies with solar elevation through the
      beam extinction coefficient (:math:`k_b`) but diffuse reflectance
      (:math:`\rho_{cd}`) is a constant property of the canopy
      (:attr:`~pyrealm.constants.two_leaf.TwoLeafConst.diffuse_reflectance`).

    * The irradiances for sunlit leaves are then calculated, including components from
      direct beam (:math:`I_{sb}`, see :meth:`calculate_sunlit_beam_irradiance`),
      scattered (:math:`I_{ss}`, see :meth:`calculate_sunlit_scattered_irradiance`) and
      diffuse (:math:`I_{sd}`, see :meth:`calculate_sunlit_diffuse_irradiance`) light.
      The total radiation absorbed by the sunlit leaves (:math:`I_csun`) is the sum of
      these three components.

    * The irradiance for shaded leaves (:math:`I_cshade`) is then calculated as the
      difference between the total canopy irradiance (:math:`I_c`) and the sunlit
      absorbed irradiance (:math:`I_csun`).

    Args:
        solar_elevation: Array of solar elevation angles (radians).
        ppfd: Array of photosynthetic photon flux density values (µmol m-2 s-1).
        leaf_area_index: Array of leaf area index values.
        patm: Array of atmospheric pressure values (pascals).
        core_constants: An instance of the core constants class.
        two_leaf_constants: An instance of the two leaf constants class.
        bounds_checker: A bounds checker instance used to validate the input data.
    """

    __experimental__ = True

    def __init__(
        self,
        solar_elevation: NDArray[np.float64],
        ppfd: NDArray[np.float64],
        leaf_area_index: NDArray[np.float64],
        patm: NDArray[np.float64],
        core_constants: CoreConst = CoreConst(),
        two_leaf_constants: TwoLeafConst = TwoLeafConst(),
        bounds_checker: BoundsChecker = BoundsChecker(),
    ):
        warn_experimental("TwoLeafIrradiance")

        # Check shapes are consistent
        check_input_shapes(solar_elevation, ppfd, leaf_area_index, patm)

        # Check bounds and store input variables
        self.solar_elevation: NDArray[np.float64] = bounds_checker.check(
            "solar_elevation", solar_elevation
        )
        r"""The solar elevation inputs (:math:`\beta`, radians)"""
        self.ppfd: NDArray[np.float64] = bounds_checker.check("ppfd", ppfd)
        """The photosynthetic photon flux density inputs (PPFD, µmol m-2 s-1)"""
        self.leaf_area_index: NDArray[np.float64] = bounds_checker.check(
            "leaf_area_index", leaf_area_index
        )
        """The leaf area index inputs (:math:`L`, unitless)"""
        self.patm: NDArray[np.float64] = bounds_checker.check("patm", patm)
        """The atmospheric pressure (:math:`P`, pascals)"""
        self.core_constants: CoreConst = core_constants
        """An instance of the core constants class."""
        self.two_leaf_constants: TwoLeafConst = two_leaf_constants
        """An instance of the two leaf constants class."""

        # Define instance attributes

        self.fraction_of_diffuse_radiation: NDArray[np.float64]
        """The fraction of diffuse radiation (:math:`f_d`)"""
        self.diffuse_irradiance: NDArray[np.float64]
        """The diffuse irradiance (:math:`I_d`) reaching the canopy."""
        self.beam_irradiance: NDArray[np.float64]
        """The beam irradiance (:math:`I_b`) reaching the canopy."""

        self.beam_extinction_coef: NDArray[np.float64]
        """The beam extinction coefficient (:math:`k_{b}`)"""
        self.scattered_beam_extinction_coef: NDArray[np.float64]
        """The scattered beam extinction coefficient (:math:`k_{b}'`)"""

        self.beam_reflectance: NDArray[np.float64]
        r"""The canopy beam reflectance for leaves with uniform angle distribution
         (:math:`\rho_{cb}`)"""

        self.sunlit_beam_irradiance: NDArray[np.float64]
        """The sunlit beam irradiance (:math:`I_{sb}`)."""
        self.sunlit_diffuse_irradiance: NDArray[np.float64]
        """The sunlit diffuse irradiance (:math:`I_{sd}`)."""
        self.sunlit_scattered_irradiance: NDArray[np.float64]
        """The sunlit scattered irradiance (:math:`I_{ss}`)."""

        self.canopy_irradiance: NDArray[np.float64]
        """The total canopy irradiance (:math:`I_c`)."""

        self.sunlit_absorbed_irradiance: NDArray[np.float64]
        """The sunlit leaf absorbed irradiance (:math:`I_{csun}`)."""
        self.shaded_absorbed_irradiance: NDArray[np.float64]
        """The shaded leaf absorbed irradiance (:math:`I_{cshade}`)."""

        # Automatically calculate the absorbed irradiances.
        self._calculate_absorbed_irradiances()

    def _calculate_absorbed_irradiances(self) -> None:
        r"""Calculate absorbed irradiance for sunlit and shaded leaves."""

        # The fraction of diffuse radiation: f_d
        self.fraction_of_diffuse_radiation = calculate_fraction_of_diffuse_radiation(
            patm=self.patm,
            solar_elevation=self.solar_elevation,
            standard_pressure=self.core_constants.k_Po,
            atmospheric_scattering=self.two_leaf_constants.atmospheric_scattering_coef,
            atmos_transmission_par=self.two_leaf_constants.atmos_transmission_par,
        )

        # The diffuse and beam irradiance reaching the canopy: I_b, I_d
        self.beam_irradiance = (1 - self.fraction_of_diffuse_radiation) * self.ppfd
        self.diffuse_irradiance = self.fraction_of_diffuse_radiation * self.ppfd

        # The beam extinction coefficient for direct light: k_b
        self.beam_extinction_coef = calculate_beam_extinction_coef(
            solar_elevation=self.solar_elevation,
            solar_obscurity_angle=self.two_leaf_constants.solar_obscurity_angle,
            extinction_numerator=self.two_leaf_constants.direct_beam_extinction_numerator,
        )

        # The extinction coefficient for scattered light: k_b'
        self.scattered_beam_extinction_coef = calculate_beam_extinction_coef(
            solar_elevation=self.solar_elevation,
            solar_obscurity_angle=self.two_leaf_constants.solar_obscurity_angle,
            extinction_numerator=self.two_leaf_constants.scattered_beam_extinction_numerator,
        )

        # The canopy beam reflectance for leaves with a uniform angle distribution:
        # rho_cb
        self.beam_reflectance = calculate_beam_reflectance(
            beam_extinction=self.beam_extinction_coef,
            horizontal_leaf_reflectance=self.two_leaf_constants.horizontal_leaf_reflectance,
        )

        # Calculate fractions of the sunlit leaf irradiance
        self.sunlit_beam_irradiance = calculate_sunlit_beam_irradiance(
            beam_irradiance=self.beam_irradiance,
            beam_extinction_coef=self.beam_extinction_coef,
            leaf_area_index=self.leaf_area_index,
            leaf_scattering_coef=self.two_leaf_constants.leaf_scattering_coef,
        )

        self.sunlit_diffuse_irradiance = calculate_sunlit_diffuse_irradiance(
            diffuse_irradiance=self.diffuse_irradiance,
            beam_extinction_coef=self.beam_extinction_coef,
            leaf_area_index=self.leaf_area_index,
            diffuse_reflectance=self.two_leaf_constants.diffuse_reflectance,
            diffuse_extinction_coef=self.two_leaf_constants.diffuse_extinction_coef,
        )

        self.sunlit_scattered_irradiance = calculate_sunlit_scattered_irradiance(
            beam_irradiance=self.beam_irradiance,
            beam_reflectance=self.beam_reflectance,
            scattered_beam_extinction_coef=self.scattered_beam_extinction_coef,
            beam_extinction_coef=self.beam_extinction_coef,
            leaf_area_index=self.leaf_area_index,
            leaf_scattering_coef=self.two_leaf_constants.leaf_scattering_coef,
        )

        # And hence the total sunlit irradiance as the sum of those three components
        self.sunlit_absorbed_irradiance = (
            self.sunlit_beam_irradiance
            + self.sunlit_diffuse_irradiance
            + self.sunlit_scattered_irradiance
        )

        # Now calculate the irradiance absorbed by shaded leaves
        # Calculate canopy irradiance
        self.canopy_irradiance = calculate_canopy_irradiance(
            beam_reflectance=self.beam_reflectance,
            beam_irradiance=self.beam_irradiance,
            scattered_beam_extinction_coef=self.scattered_beam_extinction_coef,
            diffuse_radiation=self.diffuse_irradiance,
            leaf_area_index=self.leaf_area_index,
            diffuse_reflectance=self.two_leaf_constants.diffuse_reflectance,
            diffuse_extinction_coef=self.two_leaf_constants.diffuse_extinction_coef,
        )

        self.shaded_absorbed_irradiance = np.where(
            self.solar_elevation > self.two_leaf_constants.solar_obscurity_angle,
            self.canopy_irradiance - self.sunlit_absorbed_irradiance,
            0,
        )


def calculate_beam_extinction_coef(
    solar_elevation: NDArray[np.float64],
    solar_obscurity_angle: float = TwoLeafConst().solar_obscurity_angle,
    extinction_numerator: float = TwoLeafConst().direct_beam_extinction_numerator,
) -> NDArray[np.float64]:
    r"""Calculate the beam extinction coefficient.

    The beam extinction coefficient (:math:`k_b`) captures changes in the the
    attenuation of direct sunlight through the canopy with variation in the solar
    elevation angle.

    .. math::

        k_b = \frac{n}{\sin\left(\max\left(\beta, \beta_{ob} \right)\right)}

    Args:
        solar_elevation: Array of solar elevation angles (:math:`\beta`)
        solar_obscurity_angle: Threshold angle at which the beam extinction coefficient
            reaches a maximum (:math:`\beta_{ob}`).
        extinction_numerator: The numerator used in the calculation (:math:`n`).

    Returns:
        An array of beam extinction coefficients.
    """

    return np.where(
        solar_elevation > solar_obscurity_angle,
        extinction_numerator / np.sin(solar_elevation),
        extinction_numerator / np.sin(solar_obscurity_angle),
    )


def calculate_fraction_of_diffuse_radiation(
    patm: NDArray[np.float64],
    solar_elevation: NDArray[np.float64],
    standard_pressure: float = CoreConst().k_Po,
    atmospheric_scattering: float = TwoLeafConst().atmospheric_scattering_coef,
    atmos_transmission_par: float = TwoLeafConst().atmos_transmission_par,
) -> NDArray[np.float64]:
    r"""Calculate the fraction of diffuse radiation.

    The fraction of diffuse radiation (:math:`f_d`) captures the proportion of
    sunlight that is scattered before reaching the canopy, following equation A23 and
    A25 of :cite:t:`depury:1997a`.

    .. math::
        :nowrap:

        \[
            \begin{align*}
                m &= \left(P / P_0 \right) / \sin \beta \\
                f_d &= \frac{1 - a^m}{1 + a^m \left(1 / f_a - 1\right)}
            \end{align*}
        \]


    Args:
        patm: Atmospheric pressure values (:math:`P`, pascals).
        solar_elevation: Solar elevation angles (:math:\beta`, radians).
        standard_pressure: Standard atmospheric pressure (:math:`P_0`, pascals).
        atmospheric_scattering: Atmospheric scattering factor (:math:`f_a`).
        atmos_transmission_par:  The atmospheric transmission coefficient of
            photosynthetically active radiation (:math:`a`).

    Returns:
        Array of fractions of diffuse radiation.
    """

    # Optical air mass
    m = (patm / standard_pressure) / np.sin(solar_elevation)

    # Diffuse fraction
    f_d = (1 - atmos_transmission_par**m) / (
        1 + (atmos_transmission_par**m * (1 / atmospheric_scattering - 1))
    )

    # Test for negative values - it isn't clear that this actually occurs but the
    # reference implementation clipped the resulting irradiances at zero, so this is
    # here to safeguard and inform.
    if np.any(f_d < 0):
        f_d = np.clip(a=f_d, a_min=0, a_max=None)
        warn("Negative diffuse radiation fractions clamped to zero.")

    return f_d


def calculate_beam_reflectance(
    beam_extinction: NDArray[np.float64],
    horizontal_leaf_reflectance: float = TwoLeafConst().horizontal_leaf_reflectance,
) -> NDArray[np.float64]:
    r"""Calculate the beam irradiance for leaves with a uniform angle distribution.

    The beam irradiance with a uniform leaf angle distribution (:math:`\rho_{cb}`)
    captures different leaf orientations within the canopy, following equation A19 of
    :cite:t:`depury:1997a`

    .. math::

        \rho_{cb} = 1 - \exp \left(-\frac{2 \rho_h \, k_b}{1 + k_b}\right)

    Args:
        beam_extinction: Array of beam extinction coefficients (:math:`k_b`).
        horizontal_leaf_reflectance: The reflectance coefficient horizontal leaves
            (:math:`\rho_h`).

    Returns:
        Array of beam irradiances.
    """

    return 1.0 - np.exp(
        -2 * horizontal_leaf_reflectance * beam_extinction / (1 + beam_extinction)
    )


def calculate_canopy_irradiance(
    beam_reflectance: NDArray[np.float64],
    beam_irradiance: NDArray[np.float64],
    scattered_beam_extinction_coef: NDArray[np.float64],
    diffuse_radiation: NDArray[np.float64],
    leaf_area_index: NDArray[np.float64],
    diffuse_reflectance: float = TwoLeafConst().diffuse_reflectance,
    diffuse_extinction_coef: float = TwoLeafConst().diffuse_extinction_coef,
) -> NDArray[np.float64]:
    r"""Calculate the canopy irradiance.

    The canopy irradiance (:math:`I_c`) is the total irradiance within the canopy,
    including both direct and diffuse radiation components.

    .. math::

        I_c = (1 - \rho_{cb})  I_b   (1 - \exp(-k_b' L)) +
            (1 - \rho_{cd})  I_d  (1 - \exp(-k_d' L))

    Args:
        beam_reflectance : The beam reflectance of leaves with uniform angle
            distribution  (:math:`\rho_{cb}`).
        beam_irradiance : Beam irradiance values (:math:`I_b`).
        scattered_beam_extinction_coef : Scattered beam extinction
            coefficients (:math:`k_b'`).
        diffuse_radiation : Diffuse radiation values (:math:`I_d`).
        leaf_area_index : The leaf area index  (:math:`L`).
        diffuse_reflectance : The canopy reflectance of diffuse radiation
            (:math:`\rho_{cd}`).
        diffuse_extinction_coef : The diffuse light extinction coefficients
            (:math:`k_d'`).

    Returns:
        Canopy irradiance values.
    """

    return (1 - beam_reflectance) * beam_irradiance * (
        1 - np.exp(-scattered_beam_extinction_coef * leaf_area_index)
    ) + (1 - diffuse_reflectance) * diffuse_radiation * (
        1 - np.exp(-diffuse_extinction_coef * leaf_area_index)
    )


def calculate_sunlit_beam_irradiance(
    beam_irradiance: NDArray[np.float64],
    beam_extinction_coef: NDArray[np.float64],
    leaf_area_index: NDArray[np.float64],
    leaf_scattering_coef: float = TwoLeafConst().leaf_scattering_coef,
) -> NDArray[np.float64]:
    r"""Calculate the sunlit beam irradiance.

    The sunlit beam irradiance (:math:`I_{sun_beam}`) is the direct sunlight received by
    the sunlit portion of the canopy.

    .. math::

        I_{sun\_beam} = I_b (1 - \sigma) (1 - \exp{-k_b L})

    Args:
        beam_irradiance: The irradiance by direct beam light (:math:`I_b`)
        beam_extinction_coef: Beam extinction coefficients (:math:`k_b`)
        leaf_area_index: The leaf area index for the canopy (:math:`L`)
        leaf_scattering_coef: The scattering coefficient of PAR by leaves
            (:math:`\sigma`).

    Returns:
        Array of sunlit beam irradiance values.
    """
    return (
        beam_irradiance
        * (1 - leaf_scattering_coef)
        * (1 - np.exp(-beam_extinction_coef * leaf_area_index))
    )


def calculate_sunlit_diffuse_irradiance(
    diffuse_irradiance: NDArray[np.float64],
    beam_extinction_coef: NDArray[np.float64],
    leaf_area_index: NDArray[np.float64],
    diffuse_reflectance: float = TwoLeafConst().diffuse_reflectance,
    diffuse_extinction_coef: float = TwoLeafConst().diffuse_extinction_coef,
) -> NDArray[np.float64]:
    r"""Calculate the sunlit diffuse irradiance.

    The sunlit diffuse irradiance (:math:`I_{s_d}`) is the diffuse radiation
    received by the sunlit portion of the canopy.

    .. math::

        I_{sd} = I_d  (1 - \rho_{cd}) (1 - \exp(-(k_d' + k_b)
            L \frac{k_d'}{k_d' + k_b}

    Args:
        diffuse_irradiance: Array of diffuse radiation values (:math:`I_d`)
        beam_extinction_coef: Array of beam extinction coefficients (:math:`k_b`)
        leaf_area_index: Array of leaf area index values (:math:`L`)
        diffuse_reflectance : The canopy reflectance of diffuse radiation
            (:math:`\rho_{cd}`).
        diffuse_extinction_coef: Constant for calculating the sunlit diffuse
            irradiance (:math:`k_d'`)

    Returns:
        Array of sunlit diffuse irradiance values.
    """
    return (
        diffuse_irradiance
        * (1 - diffuse_reflectance)
        * (
            1
            - np.exp(
                -(diffuse_extinction_coef + beam_extinction_coef) * leaf_area_index
            )
        )
        * diffuse_extinction_coef
        / (diffuse_extinction_coef + beam_extinction_coef)
    )


def calculate_sunlit_scattered_irradiance(
    beam_irradiance: NDArray[np.float64],
    beam_reflectance: NDArray[np.float64],
    scattered_beam_extinction_coef: NDArray[np.float64],
    beam_extinction_coef: NDArray[np.float64],
    leaf_area_index: NDArray[np.float64],
    leaf_scattering_coef: float = TwoLeafConst().leaf_scattering_coef,
) -> NDArray[np.float64]:
    r"""Calculate the sunlit scattered irradiance.

    The sunlit scattered irradiance (:math:`I_{ss}`) is the scattered
    radiation received by the sunlit portion of the canopy.

    .. math::

        I_{ss} = I_b  ((1 - \rho_{cb}) (1 - \exp(-(k_b' + k_b) L))
            \frac{k_b'}{k_b' + k_b} - (1 - \sigma) (1 - \exp(-2 k_b  L)) / 2)

    Args:
        beam_irradiance: Array of beam irradiance values (:math:`I_b`)
        beam_reflectance : The beam reflectance of leaves with uniform angle
            distribution  (:math:`\rho_{cb}`).
        scattered_beam_extinction_coef: Array of scattered beam extinction
            coefficients (:math:`k_b'`)
        beam_extinction_coef: Array of beam extinction coefficients (:math:`k_b`)
        leaf_area_index: Array of leaf area index values (:math:`L`)
        leaf_scattering_coef: The scattering coefficient of PAR by leaves
            (:math:`\sigma`).

    Returns:
        Array of sunlit scattered irradiance values.
    """

    return beam_irradiance * (
        (1 - beam_reflectance)
        * (
            1
            - np.exp(
                -(scattered_beam_extinction_coef + beam_extinction_coef)
                * leaf_area_index
            )
        )
        * scattered_beam_extinction_coef
        / (scattered_beam_extinction_coef + beam_extinction_coef)
        - (1 - leaf_scattering_coef)
        * (1 - np.exp(-2 * beam_extinction_coef * leaf_area_index))
        / 2
    )


# ------------------------------
# Assimilation class and functions
# ------------------------------


class TwoLeafAssimilation:
    r"""Estimate gross primary production using the two-leaf, two-stream model.

    The two leaf, two stream model of :cite:t:`depury:1997a` estimates irradiance within
    the canopy from diffuse, scattered and direct beam light radiation and then
    separately estimates the assimilation by sunlit and shaded leaves.

    The class requires an estimate of the irradiances absorbed by the sunlit and shaded
    leaves, calculated using the :class:`~pyrealm.pmodel.two_leaf.TwoLeafIrradiance`
    class. It also requires a :class:`~pyrealm.pmodel.pmodel.PModel` or
    :class:`~pyrealm.pmodel.pmodel.SubdailyPModel` instance, which is used to provide
    estimates of four parameters:

    * the maximum carboxylation rate (:math:`V_{cmax}`),
    * the maximum carboxylation rate at standard temperature (:math:`V_{cmax25}`),
    * the :math:`\ce{CO2}` limitation factor for RuBisCO-limited assimilation
      (:math:`m_c`), and
    * the :math:`\ce{CO2}` limitation factor for light-limited assimilation
      (:math:`m_j`).

    The model then calculate assimilation as follows:

    * An extinction coefficient (:math:`k_v`) is calculated to account for changes in
      :math:`V_{cmax}` with depth in the canopy, following an empirical function in
      :cite:t:`lloyd:2010a` (see :meth:`calculate_canopy_extinction_coef`) and using
      leaf area index (:math:`L`) as a proxy for canopy depth.

    * The value for :math:`V_{cmax25}` from the P Model is adjusted to give a
      representative through the canopy (:math:`V_{cmax25\_C}`) using :math:`L` and
      :math:`k_v` (see :meth:`calculate_canopy_vcmax25`). This carboxylation capacity is
      then partitioned between sunlit and shaded leaves: :math:`V_{cmax25\_Sn}` is
      estimated using :meth:`calculate_sun_vcmax25` and then :math:`V_{cmax25\_Sd} =
      V_{cmax25\_C} - V_{cmax25\_Sn}`.

    * Values for :math:`J_{max25}` are then calculated using an empirical function of
      :math:`V_{cmax25`}` for both sunlit (:math:`J_{max25\_Sn}`) and shaded
      (:math:`J_{max25\_Sd}`) leaves (see :meth:`calculate_jmax25`).

    * The Arrhenius scaling method used with the P Model is then used to adjust these
      estimates to the observed temperatures, giving :math:`V_{cmax\_Sn}, V_{cmax\_Sd},
      J_{max\_Sn}, J_{max\_Sd}`.

    * The maximum assimilation rate via the carboxylation pathway is then found as
      :math:`A_v = V_{cmax} m_c` for each leaf type, giving :math:`A_{v\_Sn}` and
      :math:`A_{v\_Sd}`.

    * The realised electron transport rate :math:`J` is calculated using an
      epmirical function (see :meth:`calculate_electron_transport_rate`) and then used
      to calculate the maximum assimilation rate via the electron transfer pathway as
      :math:`A_j = m_j J / 4`, giving  :math:`A_{j\_Sn}` and :math:`A_{j\_Sd}`.

    * The realised assimilation is the minimum of  :math:`A_{c}` and :math:`A_{j}` for
      each leaf type, giving:

        * :math:`A_{Sn} = \min \left( A_{v\_Sn}, A_{j\_Sn} \right)`,
        * :math:`A_{Sd} = \min \left( A_{v\_Sd}, A_{j\_Sd} \right)`.

    *  The gross primary productivity (gC m-2 s-1) is then calculated as :math:`(A_{Sn}
       + A_{Sd}) * M_C`, where :math:`M_C` is the molar mass of carbon.

    Args:
        pmodel: A PModel or SubdailyPModel.
        irrad: An instance of TwoLeafIrradiance.
    """

    __experimental__ = True

    def __init__(
        self,
        pmodel: PModel | SubdailyPModel,
        irradiance: TwoLeafIrradiance,
    ):
        """Initialize the TwoLeafAssimilation class."""

        warn_experimental("TwoLeafAssimilation")

        self.pmodel = pmodel
        """A PModel or SubdailyPModel instance."""
        self.irradiance = irradiance
        """A TwoLeafIrradiance instance """

        # TODO - both the pmodel and irrad instances have their own independent
        #        CoreConst instances. Would be good to check they are the same. Because
        #        we want to able to use the irrad object with multiple models, it would
        #        be convoluted to force them to be the same instance. We'd need to write
        #        a custom __eq__ dunder method to handle the structures inside the
        #        classes.
        #      - Similarly, both the PModel and TwoLeafIrradiance require PPFD and PATM
        #        and those should be the same between the two objects.
        #      - Ideally we want to avoid these dual definitions, which we could do if
        #        the PModel is used as input to the TwoLeafIrradiance. Less reusable
        #        class, but removes the parallel arguments.
        #      - See https://github.com/ImperialCollegeLondon/pyrealm/issues/469

        self.canopy_extinction_coef: NDArray[np.float64]
        """An extinction coefficient capturing the vertical structure of carboxylation
        capacity within the canopy (:math:`k_v`)."""
        self.Vcmax25_canopy: NDArray[np.float64]
        r"""The total canopy carboxylation capacity at standard temperature
        :math:`V_{cmax25\_C}`"""
        self.Vcmax25_sun: NDArray[np.float64]
        r"""The maximum rate of carboxylation at standard temperature within sunlit 
        leaves (:math:`V_{cmax25\_Sn}`)"""
        self.Vcmax25_shade: NDArray[np.float64]
        r"""The maximum rate of carboxylation at standard temperature within shaded 
        leaves (:math:`V_{cmax25\_Sd}`)"""
        self.Vcmax_sun: NDArray[np.float64]
        r"""The maximum rate of carboxylation at the observed temperature within sunlit
        leaves (:math:`V_{cmax\_Sn}`)"""
        self.Vcmax_shade: NDArray[np.float64]
        r"""The maximum rate of carboxylation at the observed temperature within shaded
        leaves (:math:`V_{cmax\_Sd}`)"""
        self.Jmax25_sun: NDArray[np.float64]
        r"""The maximum rate of electron transfer at standard temperature within sunlit
        leaves (:math:`J_{max25\_Sn}`)"""
        self.Jmax25_shade: NDArray[np.float64]
        r"""The maximum rate of electron transfer at standard temperature within shaded
        leaves (:math:`J_{max25\_Sd}`)"""
        self.Jmax_sun: NDArray[np.float64]
        r"""The maximum rate of electron transfer at the observed temperature within
        sunlit leaves (:math:`J_{max\_Sn}`)"""
        self.Jmax_shade: NDArray[np.float64]
        r"""The maximum rate of electron transfer at the observed temperature within
        shaded leaves (:math:`J_{max25\_Sn}`)"""
        self.J_sun: NDArray[np.float64]
        r"""The realised rate of electron transfer within sunlit leaves
        (:math:`J_{Sn}`"""
        self.J_shade: NDArray[np.float64]
        r"""The realised rate of electron transfer within sunlit leaves
        (:math:`J_{Sd}`"""
        self.Av_sun: NDArray[np.float64]
        r"""The potential rate of assimilation associated with carboxylation in sunlit
        leaves (:math:`A_{v\_Sn}`)."""
        self.Av_shade: NDArray[np.float64]
        r"""The potential rate of assimilation associated with carboxylation in shaded
        leaves (:math:`A_{v\_Sd}`)."""
        self.Aj_sun: NDArray[np.float64]
        r"""The potential rate of assimilation associated with electron transfer in
        sunlit leaves (:math:`A_{j\_Sn}`)."""
        self.Aj_shade: NDArray[np.float64]
        r"""The potential rate of assimilation associated with electron transfer in
        shaded leaves (:math:`A_{j\_Sn}`)."""
        self.A_sun: NDArray[np.float64]
        r"""The realised assimilation rate for sunlit leaves (:math:`A_{Sn}`)."""
        self.A_shade: NDArray[np.float64]
        r"""The realised assimilation rate for shaded leaves (:math:`A_{Sd}`)."""
        self.gpp: NDArray[np.float64]
        r"""The gross primary productivity across the sunlit and shaded leaves."""

        # Automatically run the
        self._calculate_two_leaf_two_stream_gpp()

    def _calculate_two_leaf_two_stream_gpp(self) -> None:
        """Internal calculations for the two leaf, two stream model."""

        # Calculate the canopy extinction coefficient given the big leaf estimate of
        # V_cmax
        self.canopy_extinction_coef = calculate_canopy_extinction_coef(
            vcmax=self.pmodel.vcmax,
            coef=self.irradiance.two_leaf_constants.vcmax_lloyd_coef,
        )

        # Calculate overall Vcmax25 for the canopy and then partition between sunlit and
        # shaded leaves
        self.Vcmax25_canopy = calculate_canopy_vcmax25(
            leaf_area_index=self.irradiance.leaf_area_index,
            vcmax25=self.pmodel.vcmax25,
            canopy_extinction_coef=self.canopy_extinction_coef,
        )

        self.Vcmax25_sun = calculate_sun_vcmax25(
            leaf_area_index=self.irradiance.leaf_area_index,
            vcmax25=self.pmodel.vcmax25,
            canopy_extinction_coef=self.canopy_extinction_coef,
            beam_extinction_coef=self.irradiance.beam_extinction_coef,
        )

        self.Vcmax25_shade = self.Vcmax25_canopy - self.Vcmax25_sun

        # Calculate Jmax25 for sunlit and shaded leaves as an empirical function of
        # Vcmax25
        self.Jmax25_sun = calculate_jmax25(self.Vcmax25_sun)
        self.Jmax25_shade = calculate_jmax25(self.Vcmax25_shade)

        # Set up Arrhenius scaling to get rates at environmental temperatures using the
        # Arrhenius method selected in the original P Model.
        arrhenius_factors = self.pmodel._arrhenius_class(env=self.pmodel.env)

        # Scale Vcmax25 to Vcmax and Jmax25 to Jmax for sunlit and shaded leaves
        arrhenius_vcmax = arrhenius_factors.calculate_arrhenius_factor(
            coefficients=self.pmodel.pmodel_const.arrhenius_vcmax
        )
        self.Vcmax_sun = self.Vcmax25_sun * arrhenius_vcmax
        self.Vcmax_shade = self.Vcmax25_shade * arrhenius_vcmax

        arrhenius_jmax = arrhenius_factors.calculate_arrhenius_factor(
            coefficients=self.pmodel.pmodel_const.arrhenius_jmax
        )
        self.Jmax_sun = self.Jmax25_sun * arrhenius_jmax
        self.Jmax_shade = self.Jmax25_shade * arrhenius_jmax

        # Calculate Av and Aj for both sun and shaded leaves
        self.Av_sun = self.Vcmax_sun * self.pmodel.optchi.mc
        self.Av_shade = self.Vcmax_shade * self.pmodel.optchi.mc

        self.J_sun = calculate_electron_transport_rate(
            jmax=self.Jmax_sun,
            absorbed_irradiance=self.irradiance.sunlit_absorbed_irradiance,
        )
        self.J_shade = calculate_electron_transport_rate(
            jmax=self.Jmax_shade,
            absorbed_irradiance=self.irradiance.shaded_absorbed_irradiance,
        )

        self.Aj_sun = self.pmodel.optchi.mj * self.J_sun / 4
        self.Aj_shade = self.pmodel.optchi.mj * self.J_shade / 4

        # Calculate the sun and shaded leaf assimilation as the minimum of Ac and Aj for
        # each component
        self.Acanopy_sun = np.minimum(self.Aj_sun, self.Av_sun)
        self.Acanopy_shade = np.minimum(self.Aj_shade, self.Av_shade)

        # Calculate GPP in gC m-2 s-1 as the sum of the sunlit and shaded components,
        # but explicitly setting to zero where the solar elevation is below the solar
        # obscurity angle.
        self.gpp = np.where(
            self.irradiance.solar_elevation
            > self.irradiance.two_leaf_constants.solar_obscurity_angle,
            (self.Acanopy_shade + self.Acanopy_sun)
            * self.irradiance.core_constants.k_c_molmass,
            0,
        )


def calculate_canopy_extinction_coef(
    vcmax: NDArray[np.float64],
    coef: tuple[float, float] = TwoLeafConst().vcmax_lloyd_coef,
) -> NDArray[np.float64]:
    r"""Calculate the canopy extinction coefficient.

    The extinction coefficient (:math:`k_v`) captures the decrease in photosynthetic
    capacity (:math:`V_{cmax}`) with depth in the plant canopy.

    The exponential model used here is taken from Figure 10 of :cite:`lloyd:2010a`,
    which presents an empirical model of the vertical profile in photosynthetic
    capacity, using data from the Amazon forest.

    .. math::
        kv_{Lloyd} = \exp(0.00963 \cdot vcmax_{pmod} - 2.43)

    Args:
        vcmax: The ``vcmax`` attribute from a P Model (:math:`V_{cmax}`).
        coef: The coefficients of the exponential momdel.

    Returns:
        NDArray: The calculated :math:`kv_Lloyd` values.
    """

    a, b = coef
    return np.exp(a * vcmax - b)


def calculate_canopy_vcmax25(
    leaf_area_index: NDArray[np.float64],
    vcmax25: NDArray[np.float64],
    canopy_extinction_coef: NDArray[np.float64],
) -> NDArray[np.float64]:
    r"""Calculate standardised carboxylation rate in the canopy.

    This function calculates the maximum carboxylation rate of the canopy at a reference
    temperature of 25°C (:math:`V_{cmax25\_C}`), given the depth of the canopy as
    estimated using leaf area index and the rate of decrease in carboxylation rate
    through the canopy, as calculated using
    :meth:`calculate_canopy_extinction_coef`.

    .. math::
        V_{cmax25\_C} = L \, V_{cmax25}  \left(\frac{1 - \exp(-k_v)}{k_v}\right)

    Args:
        leaf_area_index: The leaf area index (:math:`L`).
        vcmax25: The ``vcmax25`` parameter from a P Model (:math:`V_{cmax25}`).
        canopy_extinction_coef: The canopy extinction coefficient (:math:`k_v`).

    Returns:
        NDArray: The calculated Vmax25 canopy values.
    """
    return (
        leaf_area_index
        * vcmax25
        * ((1 - np.exp(-canopy_extinction_coef)) / canopy_extinction_coef)
    )


def calculate_sun_vcmax25(
    leaf_area_index: NDArray[np.float64],
    vcmax25: NDArray[np.float64],
    canopy_extinction_coef: NDArray[np.float64],
    beam_extinction_coef: NDArray[np.float64],
) -> NDArray[np.float64]:
    r"""Calculate standardised carboxylation rate of sunlit leaves.

    Calculates the maximum carboxylation rate for sunlit leaves at the standard
    temperature of 25°C  (:math:`V_{cmax25\_Sn}`) as:

    .. math::

        V_{cmax25\_Sn} = L \, V_{cmax25} \left(
            \frac{1 - \exp{-k_v - k_b}  L )} {k_v + k_b L}
            \right)

    Args:
        leaf_area_index: The leaf area index (LAI, :math:`L`).
        vcmax25: The ``vcmax25`` parameter from a P Model (:math:`V_{cmax25}`).
        canopy_extinction_coef: The canopy extinction coefficient (:math:`k_v`).
        beam_extinction_coef: The beam extinction coefficient from the irradiance
            model (:math:`k_b`).

    Returns:
        NDArray: The calculated Vmax25 sun values.
    """
    Vmax25_sun = (
        leaf_area_index
        * vcmax25
        * (
            (
                1
                - np.exp(
                    -canopy_extinction_coef - beam_extinction_coef * leaf_area_index
                )
            )
            / (canopy_extinction_coef + beam_extinction_coef * leaf_area_index)
        )
    )

    return Vmax25_sun


def calculate_jmax25(
    vcmax25: NDArray[np.float64],
    coef: tuple[float, float] = TwoLeafConst().jmax25_wullschleger_coef,
) -> NDArray[np.float64]:
    r"""Calculate the maximum rate of electron transport.

    This function calculates the maximum rate of electron transport (:math:`J_{max25}`)
    at 25°C as an linear function of the carboxylation rate (:math:`V_{cmax25}`),
    following the fitted model in Figure 2 of :cite:`wullschleger:1993a`. The default
    values for the coefficients (:math:`a=29,b=1.64`) come from the same source.

    .. math::

        J_{max25} = a + b V_{cmax25}

    Args:
        vcmax25: An estimate of :math:`V_{cmax25}`.
        coef: The coefficients of the empirical relationship between :math:`V_{cmax25}`
            and :math:`J_{max25}`.

    Returns:
        The calculated values of :math:`J_{max25}`.
    """

    a, b = coef
    return a + b * vcmax25


def calculate_electron_transport_rate(
    jmax: NDArray[np.float64], absorbed_irradiance: NDArray[np.float64]
) -> NDArray[np.float64]:
    r"""Calculate electron transport rate.

    This function calculates the realised electron transport rate (:math:`J`), given
    absorbed irradiance and the maximum electron transport rate :math:`J_{max}`.

    .. math::

        J = J_{max}  I_c \frac{(1 - 0.15)}{(I_c + 2.2  J_{max})}

    Args:
        jmax: The maximum rate of electron transport (:math:`J_{max}`).
        absorbed_irradiance: The abbsorbed irradiance (:math:`I_c`)

    Returns:
        The calculated J values.
    """

    # TODO  What is the source of this parameterisation?

    J = jmax * absorbed_irradiance * (1 - 0.15) / (absorbed_irradiance + 2.2 * jmax)

    return J



"""The module :mod:`~pyrealm.pmodel.pmodel_environment` provides the implementation of
the :class:`~pyrealm.pmodel.pmodel_environment.PModelEnvironment` class, which is used
to check the data required to fit a P Model and calculates the key photosynthetic
environment variables for the observations.
"""  # noqa D210, D415

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants import CoreConst, PModelConst
from pyrealm.core.bounds import BoundsChecker
from pyrealm.core.utilities import check_input_shapes, summarize_attrs
from pyrealm.pmodel.functions import (
    calc_co2_to_ca,
    calc_gammastar,
    calc_kmm,
    calc_ns_star,
)


class PModelEnvironment:
    r"""Create a PModelEnvironment instance.

    This class takes the temperature (°C), vapour pressure deficit (Pa), atmospheric
    pressure (Pa) and ambient CO2 concentration (ppm) and uses these four drivers to
    calculates four photosynthetic variables for those environmental conditions:

    * the photorespiratory :math:`\ce{CO2}` compensation point (:math:`\Gamma^{*}`,
      using :func:`~pyrealm.pmodel.functions.calc_gammastar`),
    * the relative viscosity of water (:math:`\eta^*`,
      using :func:`~pyrealm.pmodel.functions.calc_ns_star`),
    * the ambient partial pressure of :math:`\ce{CO2}` (:math:`c_a`,
      using :func:`~pyrealm.pmodel.functions.calc_co2_to_ca`) and
    * the Michaelis Menten coefficient of Rubisco-limited assimilation
      (:math:`K`, using :func:`~pyrealm.pmodel.functions.calc_kmm`).

    The ``PModelEnvironment`` will also accept values for the photosynthetic photon flux
    density (PPFD, µmol m-2 s-1) and the  fraction of absorbed photosynthetically active
    radiation (FAPAR, unitless). These values are used to calculate the absorbed
    incident radiation, which is used to scale light use efficiency up to gross primary
    productivity.

    An instance of ``PModelEnvironment`` can then be used to fit different P Models
    using the same environment but different method implementations. Note that the
    underlying constants of the P Model
    (:class:`~pyrealm.constants.pmodel_const.PModelConst`) are set when creating an
    instance of this class and will be used on all models fitted to the instance.

    In addition to the main forcing variables above, the ``PModelEnvironment`` class can
    also be used to provide additional variables used by some methods. Any additional
    arguments provided to a ``PModelEnvironment`` instance should provide an array of
    data congruent with the shape of the other forcing variables. The array dimensions
    will be checked and the argument name will be used to add the data to the instance
    as an additional attribute.

    Examples of additional variables include:

    * the volumetric soil moisture content (:math:`\theta`), required to calculate
      optimal :math:`\chi` in :meth:`~pyrealm.pmodel.optimal_chi.OptimalChiLavergne20C3`
      and :meth:`~pyrealm.pmodel.optimal_chi.OptimalChiLavergne20C3`.

    * the experimental `rootzonestress` factor used to penalise the :math:`\beta` term
      in the estimation of :math:`\chi` in
      :meth:`~pyrealm.pmodel.optimal_chi.OptimalChiPrentice14RootzoneStress` and
      :meth:`~pyrealm.pmodel.optimal_chi.OptimalChiC4RootzoneStress` and
      :meth:`~pyrealm.pmodel.optimal_chi.OptimalChiC4NoGammaRootzoneStress`.

    * The climatological aridity index, expressed as PET/P (-), used in
      :meth:`~pyrealm.pmodel.quantum_yield.QuantumYieldSandoval`.

    * The mean growth temperature, calculated as the mean temperature > 0°C during
      growing degree days (°C), also used in
      :meth:`~pyrealm.pmodel.quantum_yield.QuantumYieldSandoval`.

    .. note::

        Although the ``PModelEnvironment`` is typically used to estimate gross primary
        productivity using the P Model, not all uses require the estimation of absorbed
        incident radiation. The ``ppfd`` and ``fapar`` arguments both default to one,
        and you will need to input actual values to estimate gross primary productivity.

    Args:
        tc: Temperature, relevant for photosynthesis (°C)
        vpd: Vapour pressure deficit (Pa)
        co2: Atmospheric :math:`\ce{CO2}` concentration (ppm)
        patm: Atmospheric pressure (Pa)
        theta: Volumetric soil moisture (m3/m3)
        rootzonestress: Root zone stress factor (-)
        aridity_index: Climatological aridity index, expressed as PET/P (-)
        mean_growth_temperature: Mean growth temperature (°C)
        pmodel_const: An instance of
            :class:`~pyrealm.constants.pmodel_const.PModelConst`.
        core_const: An instance of
            :class:`~pyrealm.constants.core_const.CoreConst`.
        **kwargs: Additional data variables

    Examples:
        >>> import numpy as np
        >>> env = PModelEnvironment(
        ...     tc=np.array([20]), vpd=np.array([1000]),
        ...     co2=np.array([400]), patm=np.array([101325.0]),
        ...     fapar=np.array([1]), ppfd=np.array([800]),
        ... )
    """

    def __init__(
        self,
        tc: NDArray[np.float64],
        vpd: NDArray[np.float64],
        co2: NDArray[np.float64],
        patm: NDArray[np.float64],
        fapar: NDArray[np.float64] = np.array([1.0]),
        ppfd: NDArray[np.float64] = np.array([1.0]),
        pmodel_const: PModelConst = PModelConst(),
        core_const: CoreConst = CoreConst(),
        bounds_checker: BoundsChecker = BoundsChecker(),
        **kwargs: NDArray[np.float64],
    ):
        # Check shapes of inputs are congruent
        self.shape: tuple = check_input_shapes(
            tc, vpd, co2, patm, fapar, ppfd, *kwargs.values()
        )
        """The shape of the environmental data arrays."""

        # Validate and store the core forcing variables
        self.tc: NDArray[np.float64] = bounds_checker.check("tc", tc)
        """The temperature at which to estimate photosynthesis, °C"""
        self.vpd: NDArray[np.float64] = bounds_checker.check("vpd", vpd)
        """Vapour pressure deficit, Pa"""
        self.co2: NDArray[np.float64] = bounds_checker.check("co2", co2)
        """CO2 concentration, ppm"""
        self.patm: NDArray[np.float64] = bounds_checker.check("patm", patm)
        """Atmospheric pressure, Pa"""
        self.fapar: NDArray[np.float64] = bounds_checker.check("fapar", fapar)
        """The fraction of absorbed photosynthetically active radiation (FAPAR, -)"""
        self.ppfd: NDArray[np.float64] = bounds_checker.check("ppfd", ppfd)
        """The photosynthetic photon flux density (PPFD, µmol m-2 s-1)"""

        # Store constant settings and bounds checker
        self.pmodel_const: PModelConst = pmodel_const
        """PModel constants used to calculate environment"""
        self.core_const: CoreConst = core_const
        """Core constants used to calculate environment"""
        self._bounds_checker: BoundsChecker = bounds_checker
        """The BoundsChecker applied to the environment data."""

        # Guard against calc_density issues
        if np.nanmin(self.tc) < np.array([-25]):
            raise ValueError(
                "Cannot calculate P Model predictions for values below"
                " -25°C. See calc_density_h2o."
            )

        # Guard against negative VPD issues
        if np.nanmin(self.vpd) < np.array([0]):
            raise ValueError(
                "Negative VPD values will lead to missing data - clip to "
                "zero or explicitly set to np.nan"
            )

        # Internal calculations
        self.tk: NDArray[np.float64] = self.tc + self.core_const.k_CtoK
        """The temperature at which to estimate photosynthesis in Kelvin (K)"""

        self.ca: NDArray[np.float64] = calc_co2_to_ca(co2=self.co2, patm=self.patm)
        """Ambient CO2 partial pressure, Pa"""

        self.gammastar: NDArray[np.float64] = calc_gammastar(
            tk=self.tk,
            patm=patm,
            tk_ref=self.pmodel_const.tk_ref,
            k_Po=self.core_const.k_Po,
            coef=self.pmodel_const.bernacchi_gs,
        )
        r"""Photorespiratory compensation point (:math:`\Gamma^\ast`, Pa)"""

        self.kmm: NDArray[np.float64] = calc_kmm(
            tk=self.tk,
            patm=patm,
            tk_ref=self.pmodel_const.tk_ref,
            k_co=self.core_const.k_co,
            coef=self.pmodel_const.bernacchi_kmm,
        )
        """Michaelis Menten coefficient, Pa"""

        self.ns_star = calc_ns_star(tc=tc, patm=patm, core_const=core_const)
        """Viscosity correction factor relative to standard
        temperature and pressure, unitless"""

        # Additional variables - check bounds and add them to the instance
        for var_name, var_values in kwargs.items():
            bounds_checker.check(var_name=var_name, values=var_values)
            setattr(self, var_name, var_values)

        self._additional_vars: tuple[str, ...] = tuple(kwargs.keys())
        """A tuple containing the attribute names of additional variables passed to the
        PModelEnivronment."""

    def __repr__(self) -> str:
        """Generates a string representation of PModelEnvironment instance."""
        # DESIGN NOTE: This is deliberately extremely terse. It could contain
        # a bunch of info on the environment but that would be quite spammy
        # on screen. Having a specific summarize method that provides that info
        # is more user friendly.

        return f"PModelEnvironment(shape={self.shape})"

    def summarize(self, dp: int = 2) -> None:
        """Prints a summary of PModelEnvironment variables.

        Prints a summary of the input and photosynthetic attributes in a instance of a
        PModelEnvironment including the mean, range and number of nan values.

        Args:
            dp: The number of decimal places used in rounding summary stats.
        """

        attrs: list[tuple[str, str]] = [
            ("tc", "°C"),
            ("vpd", "Pa"),
            ("co2", "ppm"),
            ("patm", "Pa"),
            ("fapar", "-"),
            ("ppfd", "µmol m-2 s-1"),
            ("ca", "Pa"),
            ("gammastar", "Pa"),
            ("kmm", "Pa"),
            ("ns_star", "-"),
        ]

        # Add units to attribute names from bounds checker
        for this_attr in self._additional_vars:
            this_bounds = self._bounds_checker._data.get(this_attr)

            if this_bounds is not None:
                attrs.append((this_attr, this_bounds.unit))
            else:
                attrs.append((this_attr, "unknown"))

        summarize_attrs(self, tuple(attrs), dp=dp)



"""The ``splash`` submodule provides the main SplashModel class for calculating
predictions under the SPLASH model.
"""  # noqa: D205

import warnings

import numpy as np
import pandas as pd  # type: ignore
from numpy.typing import NDArray

from pyrealm.constants import CoreConst
from pyrealm.core.bounds import BoundsChecker
from pyrealm.core.calendar import Calendar
from pyrealm.core.pressure import calc_patm
from pyrealm.core.utilities import check_input_shapes
from pyrealm.splash.evap import DailyEvapFluxes
from pyrealm.splash.solar import DailySolarFluxes


class SplashModel:
    r"""Fitting the SPLASH model.

    The SplashModel class calculates the predictions of the SPLASH v1.0 model
    :cite:p:`davis:2017a`. The input variables of latitude, elevation, temperature,
    precipitation and sunshine fraction of observations are initially used to calculate
    solar and evaporative fluxes, which are stored in the ``solar`` and ``evap``
    attributes as instances of :class:`~pyrealm.splash.solar.DailySolarFluxes` and
    :class:`~pyrealm.splash.evap.DailyEvapFluxes`.

    The inputs to a SplashModel are expected to be numpy arrays with time varying along
    the first dimension. Other dimensions represent observations at sites on a
    particular date.  The ``dates`` argument is expected to be a Calendar object with
    the same length as the first dimension.

    The main use of the SplashModel object is then to calculate the expected actual
    evapotranspiration (AET), soil moisture and runoff across the time series:

    * The :meth:`~pyrealm.splash.splash.SplashModel.calculate_soil_moisture` returns
      these calculations, given an initial estimate of soil moisture in observed sites.
      This method simply iterates over the days, applying the
      :meth:`~pyrealm.splash.splash.SplashModel.estimate_daily_water_balance` method to
      calculate the daily water balance, given the soil moisture of the preceeding day.

    * The :meth:`~pyrealm.splash.splash.SplashModel.estimate_initial_soil_moisture`
      method can be used to estimate an initial soil moisture for a time series from the
      first year of data in a time series.

    Args:
        lat: The latitude of observations
        elv: The elevation of observations (m), also used to calculate atmospheric
            pressure.
        sf: The sunshine fraction (0-1, unitless)
        tc: Air temperature (°C)
        pn: Precipitation (mm/day)
        dates: The dates of the time series
        kWm: The maximum soil moisture capacity, defaulting to 150 (mm)
    """

    def __init__(
        self,
        lat: NDArray[np.float64],
        elv: NDArray[np.float64],
        sf: NDArray[np.float64],
        tc: NDArray[np.float64],
        pn: NDArray[np.float64],
        dates: Calendar,
        kWm: NDArray[np.float64] = np.array([150.0]),
        core_const: CoreConst = CoreConst(),
        bounds_checker: BoundsChecker = BoundsChecker(),
    ):
        # Check input sizes are congurent
        # TODO - think about broadcasting lat and elv rather than forcing users to do
        #        this in advance. xarray would be good here for identifying axes and
        #        checking congruence more widely.
        self.shape: tuple = check_input_shapes(elv, lat, sf, tc, pn)
        """The array shape of the input variables"""

        if len(dates) != self.shape[0]:
            raise ValueError("Number of dates must match the first dimension of inputs")

        self.elv: NDArray[np.float64] = elv
        """The elevation of sites."""
        self.lat: NDArray[np.float64] = bounds_checker.check("lat", lat)
        """The latitude of sites."""
        self.sf: NDArray[np.float64] = bounds_checker.check("sf", sf)
        """The sunshine fraction (0-1) of daily observations."""
        self.tc: NDArray[np.float64] = bounds_checker.check("tc", tc)
        """The air temperature in °C of daily observations."""
        self.pn: NDArray[np.float64] = bounds_checker.check("pn", pn)
        """The precipitation in mm of daily observations."""
        self.dates: Calendar = dates
        """The dates of observations along the first array axis."""
        self.kWm: NDArray[np.float64] = bounds_checker.check("kWm", kWm)
        """The maximum soil water capacity for sites."""

        # TODO - potentially allow _actual_ climatic pressure data as an input
        self.pa: NDArray[np.float64] = calc_patm(elv, core_const=core_const)
        """The atmospheric pressure at sites, derived from elevation"""

        # Calculate the daily solar fluxes - these are invariant across the simulation
        self.solar: DailySolarFluxes = DailySolarFluxes(
            latitude=lat,
            elevation=elv,
            dates=dates,
            sunshine_fraction=sf,
            temperature=tc,
        )
        """Estimated solar fluxes for observations"""

        # Initialise the evaporative flux class
        self.evap: DailyEvapFluxes = DailyEvapFluxes(
            solar=self.solar, pa=self.pa, tc=tc, core_const=core_const
        )
        """Estimated evaporative fluxes for observations"""

    def estimate_initial_soil_moisture(
        self,
        wn_init: NDArray[np.float64] | None = None,
        max_iter: int = 10,
        max_diff: float = 1.0,
        return_convergence: bool = False,
        verbose: bool = False,
    ) -> NDArray[np.float64]:
        """Estimate initial soil moisture.

        This method uses the first year of data provided to a SplashModel instance to
        estimate initial values for the soil moisture data. The process assumes that the
        soil moisture cycle is stationary over the first year of data and iteratively
        updates an initial guess at start of year soil moisture until those values are
        within a given tolerance of the end of year values. The method cannot be run
        when less than one year of data is provided to the model.

        The user can provide an array of initial values across sites, defaulting to an
        initial guess of zero soil moisture in all sites . The user can also control the
        maximum number of update iterations and the accepted tolerance for convergence.
        The method will normally fail if the estimates do not converge, but the
        ``return_convergence`` option can be used to return the estimated soil moisture
        at each iteration regardless of the success of convergence.

        Args:
            wn_init: An optional estimate of the start of year soil moisture.
            max_iter: The maximum number of iterations used to achieve convergence.
            max_diff: The maximum acceptable difference between year start and year end
                soil moisture,
            return_convergence: Optionally return an array of soil moistures at the end
                of each iteration even when convergence fails.
            verbose: Optionally turn on detailed logging of the iteration process.

        Returns:
            An array of the estimated starting soil moisture. If ``return_convergence``
            is set to True, the returned array will have an additional dimension for
            each iteration of the equilibration loop.

        Raises:
            ValueError: The input data are of the wrong shape, contain invalid values or
                do not include at least a full year of data.
            RuntimeError: The estimation fails to converge within the set number of
                iterations.
        """

        # Initialise loop termination
        equilibrated = False
        n_iter = 0
        wn_ret = []

        if wn_init is not None:
            # Check the shape is the same as the shape of a slice along axis 0
            if wn_init.shape != self.shape[1:]:
                raise ValueError("Incorrect shape in wn_init")
            if np.any((wn_init < 0) | (wn_init > self.kWm)):
                raise ValueError("Soil moisture must be between 0 and kWm")
            wn_start = wn_init
        else:
            wn_start = np.zeros(self.shape[1:])

        # Find a date one year into the future from the first calendar date.
        date_start = pd.Timestamp(self.dates[0].date)
        date_end = date_start + pd.DateOffset(years=1)
        num_days = (date_end - date_start).days

        if self.shape[0] < num_days:
            raise ValueError("Cannot equilibrate - less than one year of data")

        # Run the equilibration loop
        while (not equilibrated) and (n_iter < max_iter):
            # Track the iterations
            n_iter += 1

            # Loop over the calendar object, updating the soil_moisture array
            wn_day = wn_start
            for day_idx in range(num_days):
                # Calculate aet, soil moisture and runoff:
                _, wn_day, _ = self.estimate_daily_water_balance(
                    previous_wn=wn_day, day_idx=day_idx
                )

            # Calculate the difference between the start of year soil moisture and the
            # final day of the year and then update the start point to the end of year.
            diff_sm = np.abs(wn_start - wn_day)
            cur_diff = np.nanmax(diff_sm)
            wn_start = wn_day

            # Optionally store the soil moisture at the end of each loop
            if return_convergence:
                wn_ret.append(wn_start)

            # Report if verbose
            if verbose:
                print(f"Iteration: {n_iter}; maximum difference: {cur_diff}")

            if cur_diff <= max_diff:
                equilibrated = True

        # Check for convergence failure before returning the final values.
        if not equilibrated:
            msg = (
                f"Initial soil moisture did not converge within {n_iter} iterations:"
                f"maximum absolute difference = {cur_diff}"
            )
            if return_convergence:  # always returns without raising an error
                warnings.warn(msg)
            else:
                raise RuntimeError(msg)

        if return_convergence:  # returns the values of wn at each iteration
            return np.array(wn_ret)  # (n_iter, *wn_start.shape)
        else:
            return wn_start

    def estimate_daily_water_balance(
        self, previous_wn: NDArray[np.float64], day_idx: int | None = None
    ) -> tuple[NDArray, NDArray, NDArray]:
        r"""Estimate the daily water balance.

        This function estimates the daily water balance within observations. The
        function first calculates the expected actual evapotranspiration (mm d-1,
        :math:`\textrm{AET}_{[t]}`), given the soil moisture from the preceeding day
        (mm, :math:`W_{n[t-1]}`). Those are then used, along with the precipitation (mm
        d-1, :math:`P_{[t]}`) and condensation (mm d-1, :math:`C_{[t]}`) for the current
        day, to calculate the current soil moisture (mm, :math:`W_{n[t]}`) as:

        .. math::

            W_{n[t]} = W_{n[t-1]} + P_{[t]} + C_{[t]} - \textrm{AET}_{[t]}.

        When the resulting soil moisture exceeds the maximum capacity of the soil
        (``kWm``), the excess is allocated to run off, leaving the soil saturated. Note
        that the soil moisture is not altered by subsurface flow: there is not vertical
        or horizontal transfer of water from the soil, only losses through
        evapotranspiration. Negative soil moisture values are replaced by zero.

        By default, ``previous_wn`` is expected to provide estimates for all
        observations across all days in the model, but ``day_idx`` can be set to provide
        an estimate for only one particular day, for use in iterating over time series.

        Args:
            day_idx: Optionally, the index of the date for which to calculate water
                balance.
            previous_wn: Soil moisture estimates for the preceeding day (mm)

        Returns:
            A tuple of numpy arrays containing estimated  AET, daily soil moisture and
            runoff.
        """

        # Check day_idx inputs to map either the single time index given in day_idx or
        # the whole dataset.
        if day_idx is None:
            check_input_shapes(previous_wn, self.pn)
            didx: int | slice = slice(self.shape[0])
        else:
            check_input_shapes(previous_wn, self.pn[day_idx])
            didx = day_idx

        # Calculate the expected aet_d given the previous wn
        if np.any((previous_wn < 0) | (previous_wn > self.kWm)):
            raise ValueError("Soil moisture must be between 0 and kWm")

        aet = self.evap.estimate_aet(wn=previous_wn, day_idx=day_idx)

        # Calculate current soil moisture, mm
        current_wn = previous_wn + self.pn[didx] + self.evap.cond[didx] - aet

        # Partition current_wn into soil moisture and runoff (ro), mm
        # - allocate excess sm to runoff and clip out negative sm
        ro = np.clip(current_wn, self.kWm, np.inf) - self.kWm
        wn = np.clip(current_wn, 0, self.kWm)

        # Return values, ignoring the type clash that estimate_aet _can_ return
        # additional arrays. aet here is explicitly a single array not a tuple.
        return aet, wn, ro  # type: ignore

    def calculate_soil_moisture(
        self,
        wn_init: NDArray[np.float64],
    ) -> tuple[NDArray, NDArray, NDArray]:
        """Calculate the soil moisture, AET and runoff from a SplashModel.

        This function takes an initial array of soil moisture values for the first
        observations in a SplashModel time series and then iteratively applies the daily
        water balance calculations along the time axis using the
        :meth:`~pyrealm.splash.splash.SplashModel.estimate_daily_water_balance` method.
        This produces the expected actual evapotranspiration (AET), soil moisture,
        runoff and for all sites across the time series.

        Args:
            wn_init: The initial state of the soil moisture for observations

        Returns:
            A tuple of numpy arrays containing predicted AET, soil moisture and runoff.
        """

        # TODO - check input shapes

        # Create storage for outputs
        aet_out = np.full_like(self.tc, np.nan)
        wn_out = np.full_like(self.tc, np.nan)
        ro_out = np.full_like(self.tc, np.nan)

        if np.any((wn_init < 0) | (wn_init > self.kWm)):
            raise ValueError("Soil moisture must be between 0 and kWm")

        curr_wn = wn_init

        for day_idx in range(self.shape[0]):
            # Calculate the balance for this date, updating the input for
            # the following day
            aet, curr_wn, ro = self.estimate_daily_water_balance(
                curr_wn, day_idx=day_idx
            )

            # Store the outputs to return
            aet_out[day_idx] = aet
            wn_out[day_idx] = curr_wn
            ro_out[day_idx] = ro

        return aet_out, wn_out, ro_out



"""This module provides general tools for working with time series data. It currently
provides the :class:`AnnualValueCalculator`, which is used to calculate annual means and
totals of time series data.
"""  # noqa : D205

from itertools import pairwise

import numpy as np
from numpy.typing import NDArray

from pyrealm.pmodel import AcclimationModel


class AnnualValueCalculator:
    """A calculator class for annual means and totals from time series data.

    This class is used to calculate annual means and totals from time series data. An
    instance is created by providing a set of timings for the times series data, either
    as an an array of datetimes or as an AcclimationModel instance from a
    SubdailyPModel, which provides validated datetimes at subdaily temporal resolutions.

    The calculation process accounts for observations that span year boundaries, such as
    fortnightly data, by calculating the duration of each observation within each year.
    The process also handles unequal sampling intervals - such as monthly data - by
    calculating the actual duration of observations. However, with uneven sampling, the
    duration of the last interval is unknown and so an explicit endpoint must be
    provided.

    The indexing of annual subsets of observations, along with the appropriate
    weightings for observations values in calculating annual values, are calculated when
    the class is created and then used by the ``get_annual_means`` and
    ``get_annual_totals`` methods. Both methods return values for all years sampled:
    the ``year_completeness`` attribute records what fraction of a year has been sampled
    to give a particular value.

    .. Note:

        The class handles a wide range of different possible sampling frequencies and
        calculates weights for observations using the duration of observations with
        second precision. With uneven durations - such as monthly data - this will give
        slightly different values to simple means assuming equal duration.

    Example:
        >>> # Three years of monthly data
        >>> datetimes = np.arange(
        ...     np.datetime64('2000-01'),
        ...     np.datetime64('2003-01'),
        ...     np.timedelta64(1, "M")
        ... )
        >>> # Monthly data is uneven - requires an explicit endpoint.
        >>> avc = AnnualValueCalculator(datetimes, endpoint=np.datetime64('2003-01'))
        >>> avc.year_completeness
        array([1., 1., 1.])
    """

    def __init__(
        self,
        timing: AcclimationModel | NDArray[np.datetime64],
        growing_season: NDArray[np.bool_] | None = None,
        endpoint: np.datetime64 | None = None,
    ):
        # Attribute definitions
        self.datetimes: NDArray[np.datetime64]
        """The start datetime of observations taking from the initial timings"""
        self.n_obs: int
        """The number of observations in the time series."""
        self.endpoint: np.datetime64
        """A datetime giving of the end of the last observation."""
        self.growing_season: NDArray[np.bool_]
        """The initial input array of growing season data."""
        self.duration_seconds: NDArray[np.int_]
        """The duration of each observation in seconds."""
        self.indexing: list[tuple[int, int]] = []
        """Pairs of integers giving start and end indices to extract consecutive years
        of data from the time series."""
        self.duration_weights: list[NDArray[np.int_]] = []
        """A list of arrays giving the number of seconds that each observation
        within a year contributes to that year."""
        self.fractional_weights: list[NDArray[np.float64]] = []
        """A list of arrays giving the fraction of each observation within a year that
        falls in the year."""
        self.growing_season_by_year: list[NDArray[np.bool_]]
        """A list of arrays giving the growing season subarrays for each year."""
        self.year_completeness: NDArray[np.float64]
        """Provides the fractional coverage of observations for each year."""
        self.years: NDArray[np.datetime64]
        """The covered years as np.datetime64 at year precision."""
        self.year_total_seconds: NDArray[np.int_]
        """The total number of seconds for each year in the time series."""
        self.year_n_days: NDArray[np.floating]
        """The total number of days in each year in the time series."""
        self.year_n_growing_days: NDArray[np.floating]
        """The total number of growing days for each year in the time series. If the
        growing_season input varies within days, these values can contain non-integer
        values."""

        # Sanity checks on datetimes
        if not (
            isinstance(timing, AcclimationModel)
            or (
                isinstance(timing, np.ndarray)
                and np.issubdtype(timing.dtype, np.datetime64)
            )
        ):
            raise ValueError(
                "The timings argument must be an AcclimationModel "
                "or an array of datetime64 values"
            )

        if isinstance(timing, AcclimationModel):
            # AcclimationModel by construction provides subdaily data with equal spacing
            self.datetimes = timing.datetimes.astype("datetime64[s]")
            duration_last_observation = timing.spacing.astype("timedelta64")
        else:
            # Pure datetime inputs could be any frequency from subdaily to monthly, and
            # some frequencies could be of differing lengths (monthly being a good
            # example)

            # Convert time to seconds precision
            self.datetimes = timing.astype("datetime64[s]")

            # Get the intervals in seconds and see if they are strictly increasing and
            # then if the spaing is consistent. If the spacing is not consistent, then
            # require an endpoint for the observations.
            duration_seconds = np.diff(self.datetimes)

            if not np.all(duration_seconds > 0):
                raise ValueError("The timing values are not strictly increasing")

            intervals: NDArray = np.unique(duration_seconds)

            if len(intervals) == 1:
                # Constant intervals
                duration_last_observation = duration_seconds[0]
            else:
                if endpoint is None:
                    raise ValueError(
                        "The timings values are not equally spaced: provide an "
                        "explicit endpoint"
                    )

                if endpoint <= timing[-1]:
                    raise ValueError(
                        "The end_datetime value must be greater than the "
                        "last timing value"
                    )

                duration_last_observation = (endpoint - self.datetimes[-1]).astype(
                    "timedelta64[s]"
                )

        self.n_obs = self.datetimes.size

        # Sanity checks on growing season
        if growing_season is None:
            growing_season = np.ones_like(self.datetimes, dtype=np.bool_)
        else:
            if not np.issubdtype(growing_season.dtype, np.bool_):
                raise ValueError(
                    "Growing season data is not an array of boolean values"
                )

            if not self.datetimes.shape == growing_season.shape:
                raise ValueError(
                    "Growing season data is not the same shape as the timing data"
                )
        # Store the growing season data
        self.growing_season = growing_season

        # Record the endpoint to get the total timespan of the data and hence the
        # duration of each observation
        self.endpoint = self.datetimes[-1] + duration_last_observation
        timespan = np.append(self.datetimes, self.endpoint)
        self.duration_seconds = np.diff(timespan).astype(np.int_)

        # Now get the datetimes of the start of each year included in the data
        years = np.unique(timespan.astype("datetime64[Y]"))

        # Unless the last timespan value is exactly equal to the end of the previous
        # year, add the next year to the list of years to handle trailing data.
        # Also record the identity of each year in a way that handles years ending
        # exactly on the year change (and which therefore end with a year that has no
        # data)
        if not (years[-1] == timespan[-1]):
            self.years = np.copy(years)
            years = np.append(years, years[-1] + np.timedelta64(1, "Y"))
        else:
            self.years = np.copy(years[:-1])

        # Convert to second precision and find where they occur in the timespan
        years = years.astype("datetime64[s]")
        year_change_indices = np.searchsorted(timespan, years)

        # Now assign the duration of each observation across years, allowing for year
        # changes that occur during an observation, storing the indices of subsets and
        # the weighting to be used with values.

        # Iterate over pairs of year dates and indices
        for (lower, upper), (lower_index, upper_index) in zip(
            pairwise(years), pairwise(year_change_indices)
        ):
            # Get the initial set of datetimes within the year
            year_datetimes = timespan[lower_index:upper_index]

            # If the upper index is not to the end of the time series, then append the
            # closing time for the current year at the end and extend the sample to
            # include the next value.
            #
            # Note here that the indexing of the final observation does not require
            # special handling because np.searchsorted returns a last index _beyond_ the
            # end of the timespan, so will automatically include the last observation.
            if upper_index < len(timespan):
                year_datetimes = np.append(year_datetimes, upper)

            # If the first observation is not the precise start of the year _and_ we are
            # not on the first year of data, then we also need to shift lower_index down
            # to include partial data from the previous observation and add the year
            # start to the internal datetimes
            if (year_datetimes[0] != lower) and (lower_index > 0):
                lower_index -= 1
                year_datetimes = np.insert(year_datetimes, 0, lower)

            # Calculate the duration of the observations within the year span
            internal_year_durations = np.diff(year_datetimes).astype(np.int_)

            # Divide the internal duration through by the actual observation durations
            # to get fractional weights.
            fractional_duration = (
                internal_year_durations / self.duration_seconds[lower_index:upper_index]
            )

            # Store the indices and weights
            self.indexing.append((int(lower_index), int(upper_index)))
            self.duration_weights.append(internal_year_durations)
            self.fractional_weights.append(fractional_duration)

        # Split the growing season up into a list of subarrays by year
        self.growing_season_by_year = [
            growing_season[lower:upper] for lower, upper in self.indexing
        ]

        # Populate the year completeness and day counts
        self.year_total_seconds = np.diff(years).astype(np.int_)
        self.year_completeness = (
            np.array([np.sum(v) for v in self.duration_weights])
            / self.year_total_seconds
        )
        day_seconds = 86400
        self.year_n_days = self.year_total_seconds / day_seconds
        self.year_n_growing_days = np.array(
            [
                np.sum(x * y) / day_seconds
                for x, y in zip(self.duration_weights, self.growing_season_by_year)
            ]
        )

    def _split_values_by_year(
        self, values: NDArray[np.float64]
    ) -> list[NDArray[np.float64]]:
        """Validates and splits value arrays.

        Args:
            values: An array of values.
        """

        if values.shape[0] != self.n_obs:
            raise ValueError(
                "First axis of values shape does not match number of observations."
            )

        # Split the daily values into subarrays for each year
        return [values[lower:upper] for lower, upper in self.indexing]

    def get_annual_means(
        self,
        values: NDArray[np.float64],
        within_growing_season: bool = False,
    ) -> NDArray[np.floating]:
        """Get annual means from an array of values.

        Average values are calculated weighted by the __duration__ of each observation,
        including weighting partial observations than span year boundaries. If
        ``within_growing_season`` is ``True``, the weights for observations outside of
        the observations marked as the growing season are set to zero. The calculation
        handles missing values.

        Example:
            >>> # Three years of monthly data
            >>> datetimes = np.arange(
            ...     np.datetime64('2000-01'),
            ...     np.datetime64('2003-01'),
            ...     np.timedelta64(1, "M")
            ... )
            >>> # Monthly data is uneven - requires an explicit endpoint.
            >>> avc = AnnualValueCalculator(
            ...     datetimes, endpoint=np.datetime64('2003-01')
            ... )
            >>> # Note that the means are weighted by the actual durations of months.
            >>> avc.get_annual_means(np.arange(0, 36)).round(4)
            array([ 5.5137, 17.526 , 29.526 ])

        Args:
            values: The data to summarize by year
            within_growing_season: Should the mean only include values within the
                growing season.
        """

        values_by_year = self._split_values_by_year(values)

        # Averages use _duration_ weights
        if within_growing_season:
            weights = [
                wght * gs
                for wght, gs in zip(self.duration_weights, self.growing_season_by_year)
            ]
        else:
            weights = self.duration_weights

        # Calculate the weighted mean in a np.nan friendly way: the product of np.nan
        # and a weight is np.nan and the isnan term omits the weights of nan
        # observations from the weighted average.
        return np.array(
            [
                np.nansum(vals * wghts) / np.nansum(~np.isnan(vals) * wghts)
                for vals, wghts in zip(values_by_year, weights)
            ]
        )

    def get_annual_totals(
        self,
        values: NDArray[np.float64],
        within_growing_season: bool = False,
    ) -> NDArray[np.floating]:
        """Get annual totals from an array of values.

        The contribution of each observation to the total is weighted by the
        __fractional__ duration of each observation within each year in order to
        partition the sum across years correctly. If ``within_growing_season`` is
        ``True``, the weights for observations not identified as growing season values
        are set to zero. The method handles missing data (`np.nan`) but obviously the
        resulting annual total will be reduced.

        Example:
            >>> # Three years of monthly data with incomplete years at start and end
            >>> datetimes = np.arange(
            ...     np.datetime64('2000-07'),
            ...     np.datetime64('2003-07'),
            ...     np.timedelta64(1, "M")
            ... )
            >>> # Monthly data is uneven - requires an explicit endpoint.
            >>> avc = AnnualValueCalculator(
            ...     datetimes, endpoint=np.datetime64('2003-07')
            ... )
            >>> # Note that the means are weighted by the actual durations of months.
            >>> avc.get_annual_totals(np.arange(0, 36)).round(4)
            array([ 15., 138., 282., 195.])
            >>> # Year completeness: 184/366 days in 2000, 181/365 days in 2003.
            >>> avc.year_completeness.round(4)
            array([0.5027, 1.    , 1.    , 0.4959])

        Args:
            values: The data to summarize by year
            within_growing_season: Should the mean only include values within the
                growing season.
        """

        values_by_year = self._split_values_by_year(values)

        # Totals use _fractional_ weights
        if within_growing_season:
            weights = [
                wght * gs
                for wght, gs in zip(
                    self.fractional_weights, self.growing_season_by_year
                )
            ]
        else:
            weights = self.fractional_weights

        return np.array(
            [np.nansum(vals * wghts) for vals, wghts in zip(values_by_year, weights)]
        )



"""Functionality for canopy modelling."""

from __future__ import annotations

from dataclasses import InitVar, dataclass, field
from typing import ClassVar

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import root_scalar  # type: ignore [import-untyped]

from pyrealm.core.experimental import warn_experimental
from pyrealm.demography.community import Community
from pyrealm.demography.core import PandasExporter, _validate_demography_array_arguments
from pyrealm.demography.crown import (
    CrownProfile,
    calculate_relative_crown_radius_at_z,
    calculate_stem_projected_crown_area_at_z,
)


def solve_canopy_area_filling_height(
    z: float,
    stem_height: NDArray[np.float64],
    crown_area: NDArray[np.float64],
    m: NDArray[np.float64],
    n: NDArray[np.float64],
    q_m: NDArray[np.float64],
    z_max: NDArray[np.float64],
    n_individuals: NDArray[np.float64],
    target_area: float = 0,
    validate: bool = True,
) -> NDArray[np.float64]:
    """Solver function for finding the height where a canopy occupies a given area.

    This function takes the number of individuals in each cohort along with the stem
    height and crown area and a given vertical height (:math:`z`). It then uses the
    crown shape parameters associated with each cohort to calculate the community wide
    projected crown area above that height (:math:`A_p(z)`). This is simply the sum of
    the products of the individual stem crown projected area at :math:`z` and the number
    of individuals in each cohort.

    The return value is the difference between the calculated :math:`A_p(z)` and a
    user-specified target area, This allows the function to be used with a root solver
    to find :math:`z` values that result in a given :math:`A_p(z)`. The default target
    area is zero, so the default return value will be the actual total :math:`A_p(z)`
    for the community.

    A typical use case for the target area would be to specify the area at which a given
    canopy layer closes under the perfect plasticity approximation in order to find the
    closure height.

    Args:
        z: Vertical height on the z axis.
        n_individuals: Number of individuals in each cohort
        crown_area: Crown area of each cohort
        stem_height: Stem height of each cohort
        m: Crown shape parameter ``m``` for each cohort
        n: Crown shape parameter ``n``` for each cohort
        q_m: Crown shape parameter ``q_m``` for each cohort
        z_max: Crown shape parameter ``z_m``` for each cohort
        target_area: A target projected crown area.
        validate: Boolean flag to suppress argument validation.
    """
    # Convert z to array for validation and typing
    z_arr = np.array(z)

    if validate:
        _validate_demography_array_arguments(
            trait_args={"m": m, "n": n, "q_m": q_m, "n_individuals": n_individuals},
            size_args={
                "z": z_arr,
                "crown_area": crown_area,
                "stem_height": stem_height,
                "z_max": z_max,
            },
        )

    q_z = calculate_relative_crown_radius_at_z(
        z=z_arr, stem_height=stem_height, m=m, n=n, validate=False
    )
    # Calculate A(p) for the stems in each cohort
    A_p = calculate_stem_projected_crown_area_at_z(
        z=z_arr,
        q_z=q_z,
        stem_height=stem_height,
        crown_area=crown_area,
        q_m=q_m,
        z_max=z_max,
        validate=False,
    )

    return (A_p * n_individuals).sum() - target_area


def fit_perfect_plasticity_approximation(
    community: Community,
    canopy_gap_fraction: float,
    max_stem_height: float,
    solver_tolerance: float,
) -> NDArray[np.float64]:
    r"""Find canopy layer heights under the PPA model.

    Finds the closure heights of the canopy layers under the perfect plasticity
    approximation by fidnding the set of heights that lead to complete closure of canopy
    layers through the canopy. The function solves the following equation for integers
    :math:`l \in (1,2,..., m)`:

    .. math::

        \sum_{s=1}^{N_s}{ A_p(z^*_l)} = l A(1 - f_G)

    The right hand side sets out the total area needed to close a given layer :math:`l`
    and all layers above it:  :math:`l` times the total community area  :math:`A` less
    any canopy gap fraction (:math:`f_G`). The left hand side then calculates the
    projected crown area for each stem :math:`s` :math:`A_p(z^*_l)_{[s]}` and sums those
    areas across all stems in the community  :math:`N_s`. The specific height
    :math:`z^*_l` is then the height at which the two terms are equal and hence solves
    the equation for layer :math:`l`.

    Args:
        community: A community instance providing plant cohort data
        canopy_gap_fraction: The canopy gap fraction
        max_stem_height: The maximum stem height in the canopy, used as an upper bound
            on finding the closure height of the topmost layer.
        solver_tolerance: The absolute tolerance used with the root solver to find the
            layer heights.
    """

    # Calculate the number of layers to contain the total community crown area
    total_community_crown_area = (
        community.stem_allometry.crown_area * community.cohorts.n_individuals
    ).sum()
    crown_area_per_layer = community.cell_area * (1 - canopy_gap_fraction)
    n_layers = int(np.ceil(total_community_crown_area / crown_area_per_layer))

    # Initialise the layer heights array and then loop over the layers indices,
    # except for the final layer, which will be the partial remaining vegetation below
    # the last closed layer.
    layer_heights = np.zeros(n_layers, dtype=np.float64)
    upper_bound = max_stem_height

    for layer in np.arange(n_layers - 1):
        # Set the target area for this layer
        target_area = (layer + 1) * crown_area_per_layer

        # TODO - the solution is typically closer to the upper bound of the bracket,
        # there might be a better algorithm to find the root (#293).
        solution = root_scalar(
            solve_canopy_area_filling_height,
            args=(
                community.stem_allometry.stem_height,
                community.stem_allometry.crown_area,
                community.stem_traits.m,
                community.stem_traits.n,
                community.stem_traits.q_m,
                community.stem_allometry.crown_z_max,
                community.cohorts.n_individuals,
                target_area,
                False,  # validate
            ),
            bracket=(0, upper_bound),
            xtol=solver_tolerance,
        )

        if not solution.converged:
            raise RuntimeError(
                "Estimation of canopy layer closure heights failed to converge."
            )

        # Store the solution and update the upper bound for the next layer down.
        layer_heights[layer] = upper_bound = solution.root

    return layer_heights[:, None]


@dataclass
class CohortCanopyData(PandasExporter):
    """Dataclass holding canopy data across cohorts.

    The cohort canopy data consists of a set of attributes represented as two
    dimensional arrays. Each row is different height at which canopy properties are
    required and the columns represent the different cohorts or the identical stem
    properties of individuals within cohorts.

    The data class:

    1. Takes the projected leaf area at the required heights and then partitions this
       into the actual leaf area within each layer, the leaf area index across the whole
       cohort and then then light absorption and transmission fractions of each cohort
       at each level.

    2. Calculates the community-wide transmission and absorption profiles. These are
       generated as an instance of the class
       :class:`~pyrealm.demography.canopy.CommunityCanopyData` and stored in the
       ``community_data`` attribute.

    3. Allocates the community-wide absorption across cohorts. The total fraction of
       light absorbed across layers is a community-wide property
       - each cohort contributes to the cumulative light absorption. Once the light
       absorbed within a layer of the community is known, this can then be partitioned
       back to cohorts and individual stems to give the fraction of canopy top
       radiation intercepted by each stem within each layer.

    Args:
        projected_leaf_area: A two dimensional array providing projected leaf area for a
            set of cohorts (columns) at a set of required heights (rows), as for example
            calculated using the :class:`~pyrealm.demography.crown.CrownProfile` class.
        n_individuals: A one-dimensional array of the number of individuals in each
            cohort.
        lai: A one-dimensional array giving the leaf area index trait for the plant
            functional type of each cohort.
        par_ext: A one-dimensional array giving the light extinction coefficient for
            the plant functional type of each cohort.
        cell_area: A float setting the total canopy area available to the cohorts.
    """

    array_attrs: ClassVar[tuple[str, ...]] = (
        "stem_leaf_area",
        "fapar",
    )

    # Init vars
    projected_leaf_area: InitVar[NDArray[np.float64]]
    """An array of the stem projected leaf area for each cohort at each of the required
    heights."""
    n_individuals: InitVar[NDArray[np.int_]]
    """The number of individuals for each cohort."""
    lai: InitVar[NDArray[np.float64]]
    """The leaf area index of the plant functional type for each cohort."""
    par_ext: InitVar[NDArray[np.float64]]
    """The extinction coefficient of the plant functional type for each cohort."""
    cell_area: InitVar[float]
    """The area available to the community."""

    # Computed variables
    stem_leaf_area: NDArray[np.float64] = field(init=False)
    """The leaf area of the crown model for each cohort by layer."""
    cohort_absorption: NDArray[np.float64] = field(init=False)
    """The Beer-Lambert absorption fraction for each cohort."""
    fapar: NDArray[np.floating] = field(init=False)
    """The across layer fractions of absorbed radiation for each cohort by layer."""

    # Community wide attributes in their own class
    community_data: CommunityCanopyData = field(init=False)
    """The community wide canopy properties."""

    __experimental__ = True

    def __post_init__(
        self,
        projected_leaf_area: NDArray[np.float64],
        n_individuals: NDArray[np.int_],
        lai: NDArray[np.float64],
        par_ext: NDArray[np.float64],
        cell_area: float,
    ) -> None:
        """Calculates cohort canopy attributes from the input data."""

        warn_experimental("CohortCanopyData")

        # Partition the projected leaf area into the leaf area in each layer for each
        # stem and then scale up to the cohort leaf area in each layer.
        self.stem_leaf_area = np.diff(projected_leaf_area, axis=0, prepend=0)

        # Calculate Beer-Lambert light absorption coefficient for each cohort
        self.cohort_absorption = 1 - np.exp(-par_ext * lai)

        # Calculate the community wide properties across cohorts:
        # - average light transmission through each layer
        # - average lai within each layer
        # - the cumulative light transmission profile
        self.community_data = CommunityCanopyData(
            absorption=self.cohort_absorption,
            leaf_area_index=lai,
            cohort_leaf_area=self.stem_leaf_area * n_individuals,
            cell_area=cell_area,
        )

        # Now calculate the theoretical fAPAR for each cohort in each layer as the outer
        # product of the within layer absorption for each cohort and the transmission
        # profile giving the mean light transmission through each layer.
        self.fapar = np.outer(
            self.community_data.transmission_profile, self.cohort_absorption
        )


@dataclass
class CommunityCanopyData(PandasExporter):
    """Dataclass holding community-wide canopy data.

    The community canopy data consists of a set of attributes represented as one
    dimensional arrays, with each entry representing a different vertical height at
    which canopy properties are required.

    The data class takes the expected light transmission for each cohort within each
    layer (as the prediction from the Beer-Lambert law for the cohort), along with the
    total leaf area in each layer within each cohort, and uses this to calculate the
    average light transmission profile down through the canopy layers. It also
    calculates the average leaf area index within each layer.

    The cumulative transmission profile shows the fraction of light reaching each of the
    canopy layers given the average absorption of the layer above, starting with 1 to
    represent the light reaching the canopy top. The fraction of light reaching the
    ground below the canopy is stored as the `transmission_to_ground` attribute.

    Args:
        layer_absorption: The expected light absorption for cohorts within each layer.
        layer_leaf_area_index: The leaf area index of cohorts within layers.
        cohort_leaf_area: The total leaf area of each cohort in each layer.
        cell_area: The area of the cell containing the community.
    """

    array_attrs: ClassVar[tuple[str, ...]] = (
        "average_layer_absorption",
        "average_layer_lai",
        "transmission_profile",
    )

    # Init vars
    absorption: InitVar[NDArray[np.float64]]
    """The Beer Lambert light absorption fraction for each cohort."""
    leaf_area_index: InitVar[NDArray[np.float64]]
    """The leaf area index for each cohort."""
    cohort_leaf_area: InitVar[NDArray[np.float64]]
    """The total leaf area per cohort for each layer."""
    cell_area: InitVar[float]
    """The total area within the community."""

    # Calculated variables
    average_layer_absorption: NDArray[np.float64] = field(init=False)
    """The average absorption within layers across the community."""
    average_layer_lai: NDArray[np.float64] = field(init=False)
    """The average leaf area index of the community within layers."""
    transmission_profile: NDArray[np.float64] = field(init=False)
    """The light transmission profile through the canopy by layer."""
    transmission_to_ground: NDArray[np.float64] = field(init=False)
    """The fraction of light reaching the ground below the canopy."""

    __experimental__ = True

    def __post_init__(
        self,
        absorption: NDArray[np.float64],
        leaf_area_index: NDArray[np.float64],
        cohort_leaf_area: NDArray[np.float64],
        cell_area: float,
    ) -> None:
        """Calculates community-wide canopy attributes from the input data."""

        warn_experimental("CommunityCanopyData")

        # Get the average absorption within the layer: per cohort values weighted by
        # cohort leaf area and scaled back to cell area.
        self.average_layer_absorption = (absorption * cohort_leaf_area).sum(
            axis=1
        ) / cell_area

        # Calculate the average LAI within the layer
        self.average_layer_lai = (leaf_area_index * cohort_leaf_area).sum(
            axis=1
        ) / cell_area

        # Calculate the cumulative light transmission profile across layers, adding 1 to
        # give the proportion of light available to the first layer.
        full_transmission = np.cumprod(
            np.concat([[1], (1 - self.average_layer_absorption)])
        )
        self.transmission_profile = full_transmission[:-1]
        self.transmission_to_ground = full_transmission[-1]


class Canopy:
    """Calculate canopy characteristics for a plant community.

    This class generates a canopy structure for a community of trees using the
    perfect-plasticity approximation (PPA) model :cite:`purves:2008a`. In this approach,
    each individual is assumed to arrange its canopy crown area plastically to take up
    space in canopy layers and that new layers form below the canopy top as the
    available space is occupied.

    Real canopies contain canopy gaps, through process such as crown shyness. This
    is included in the model through the canopy gap fraction, which sets the proportion
    of the available space that will remain unfilled by any canopy.

    Args:
        community: A Community object that will be used to generate the canopy model.
        layer_heights: A column array of vertical heights at which to calculate canopy
            variables.
        fit_ppa: Calculate layer heights as the canopy layer closure heights under the
            PPA model.
        canopy_gap_fraction: The proportion of the available space unfilled by canopy
            (default: 0.05).
        layer_tolerance: The minimum precision used by the solver to find canopy layer
            closure heights (default: 0.001 metres)
    """

    __experimental__ = True

    def __init__(
        self,
        community: Community,
        layer_heights: NDArray[np.float64] | None = None,
        fit_ppa: bool = False,
        canopy_gap_fraction: float = 0,
        solver_tolerance: float = 0.001,
    ) -> None:
        # Warn that this is an experimental feature
        warn_experimental("Canopy")

        # Store required init vars
        self.canopy_gap_fraction: float = canopy_gap_fraction
        """Canopy gap fraction."""
        self.solver_tolerance: float = solver_tolerance
        """Numerical tolerance for fitting the PPA model of canopy layer closure."""

        # Define class attributes
        self.max_stem_height: float
        """Maximum height of any individual in the community (m)."""
        self.n_layers: int
        """Total number of canopy layers."""
        self.n_cohorts: int
        """Total number of cohorts in the canopy."""
        self.heights: NDArray[np.float64]
        """The vertical heights at which the canopy structure is calculated."""

        self.crown_profile: CrownProfile
        """The crown profiles of the community stems at the provided layer heights."""
        self.cohort_data: CohortCanopyData
        """The per-cohort canopy data."""
        self.community_data: CommunityCanopyData
        """The community-wide canopy data."""
        self.filled_community_area: float
        """The area filled by crown after accounting for the crown gap fraction."""

        # Check operating mode
        if fit_ppa ^ (layer_heights is None):
            raise ValueError("Either set fit_ppa=True or provide layer heights.")

        # Set simple attributes
        self.max_stem_height = community.stem_allometry.stem_height.max()
        self.n_cohorts = community.n_cohorts
        self.filled_community_area = community.cell_area * (
            1 - self.canopy_gap_fraction
        )

        # Populate layer heights
        if layer_heights is not None:
            self.heights = layer_heights
        else:
            self.heights = fit_perfect_plasticity_approximation(
                community=community,
                canopy_gap_fraction=canopy_gap_fraction,
                max_stem_height=self.max_stem_height,
                solver_tolerance=solver_tolerance,
            )

        self._calculate_canopy(community=community)

    def _calculate_canopy(self, community: Community) -> None:
        """Calculate the canopy structure.

        This private method runs the calculations needed to populate the instance
        attributes, given the layer heights provided by the user or calculated using the
        PPA model.

        Args:
            community: The Community object passed to the instance.
        """

        # Calculate the crown profile at the layer heights
        # TODO - reimpose validation
        self.crown_profile = CrownProfile(
            stem_traits=community.stem_traits,
            stem_allometry=community.stem_allometry,
            z=self.heights,
        )

        # Calculate the per cohort canopy components (LAI, f_trans, f_abs) from the
        # projected leaf area for each stem at the layer heights
        self.cohort_data = CohortCanopyData(
            projected_leaf_area=self.crown_profile.projected_leaf_area,
            n_individuals=community.cohorts.n_individuals,
            lai=community.stem_traits.lai,
            par_ext=community.stem_traits.par_ext,
            cell_area=community.cell_area,
        )

        # Create a shorter reference to the community data
        self.community_data = self.cohort_data.community_data
